
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "3dmode.h"

extern bool gravity;


Camera::Camera()
{
	position.set(0.0f, 0.0f, 0.0f);
	view.set(0.0f, 1.0f, 0.5f);
	up_vector.set(0.0f, 0.0f, 1.0f);
}

void Camera::PositionCamera(float pos_x,	float pos_y,	float pos_z,
			 				float view_x,	float view_y,	float view_z,
							float up_x,		float up_y,		float up_z)
{
	position.set(pos_x, pos_y, pos_z);
	view.set(view_x, view_y, view_z);
	up_vector.set(up_x, up_y, up_z);
}

void Camera::RotateView(float angle, float x, float y, float z)
{
	point3_t new_view;

	// Get the view vector (The direction we are facing)
	point3_t v_view = view - position;		

	// Calculate the sine and cosine of the angle once
	float cosTheta = (float)cos(angle);
	float sinTheta = (float)sin(angle);

	// Find the new x position for the new rotated point
	new_view.x  = (cosTheta + (1 - cosTheta) * x * x)		* v_view.x;
	new_view.x += ((1 - cosTheta) * x * y - z * sinTheta)	* v_view.y;
	new_view.x += ((1 - cosTheta) * x * z + y * sinTheta)	* v_view.z;

	// Find the new y position for the new rotated point
	new_view.y  = ((1 - cosTheta) * x * y + z * sinTheta)	* v_view.x;
	new_view.y += (cosTheta + (1 - cosTheta) * y * y)		* v_view.y;
	new_view.y += ((1 - cosTheta) * y * z - x * sinTheta)	* v_view.z;

	// Find the new z position for the new rotated point
	new_view.z  = ((1 - cosTheta) * x * z - y * sinTheta)	* v_view.x;
	new_view.z += ((1 - cosTheta) * y * z + x * sinTheta)	* v_view.y;
	new_view.z += (cosTheta + (1 - cosTheta) * z * z)		* v_view.z;

	// Now we just add the newly rotated vector to our position to set
	// our new rotated view of our camera.
	view = position + new_view;
}


void Camera::StrafeCamera(float speed)
{	
	// Add the strafe vector to our position
	position.x += strafe.x * speed;
	position.y += strafe.y * speed;

	// Add the strafe vector to our view
	view.x += strafe.x * speed;
	view.y += strafe.y * speed;
}

void Camera::MoveCamera(float speed)
{
	// Get the current view vector (the direction we are looking)
	point3_t vector = view - position;
	vector.set(vector.normalize());

	position.x += vector.x * speed;		// Add our acceleration to our position's X
	position.y += vector.y * speed;		// Add our acceleration to our position's Y

	if (!gravity)
		position.z += vector.z * speed;		// Add our acceleration to our position's Z

	view.x += vector.x * speed;			// Add our acceleration to our view's X
	view.y += vector.y * speed;			// Add our acceleration to our view's Y

	if (!gravity)
		view.z += vector.z * speed;			// Add our acceleration to our view's Z
}

void Camera::Update() 
{
	// Initialize a variable for the cross product result
	point3_t v_cross = cross(view - position, up_vector);

	// Normalize the strafe vector
	strafe = v_cross.normalize();
}

void Camera::Look()
{
	// Give openGL our camera position, then camera view, then camera up vector
	gluLookAt(position.x, position.y, position.z,	
			  view.x,	 view.y,     view.z,	
			  up_vector.x, up_vector.y, up_vector.z);
}