// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004                  >>
// << ------------------------------------ >>
// << input.cpp - Input handling functions >>
// << ------------------------------------ >>

// DEFINES ------------------------------- >>
#define KEY_3D_DELAY 7

// INCLUDES ------------------------------ >>
#include "main.h"
#include "struct_3d.h"
#include "camera.h"
#include "menu.h"
#include "keybind.h"
#include "gui.h"
#include "map.h"
#include "tex_browse.h"
#include "3dmode.h"
#include "console.h"
#include "undoredo.h"
#include "edit.h"

// VARIABLES ----------------------------- >>
Uint8*		keys;
bool		gravity = true;
int			key_3d_rep = 0;

CVAR(Float, move_speed_3d, 0.05f, CVAR_SAVE)

// EXTERNAL VARIABLES -------------------- >>
extern BindList binds;
extern int xoff, yoff;
extern float zoom;
extern int edit_mode, map_changed;
extern int gridsize;
extern point2_t	mouse;
extern rect_t sel_box;

extern SDL_Surface *screen;
extern Camera		camera;

extern bool mode_3d, console;

extern BYTE	sel_item, max_item, open_menu, n_columns;
extern bool menu, inbox, inbox_typed, line_draw, lock_hilight;
extern string menu_input, browser_search;

extern Map map;
extern float scale, console_height;
extern string console_command;

extern string msg_text;
extern int	msg_count;

extern int hilight_item;

extern stringlist_t line_options, thing_options, sector_options, edit_menu;

EXTERN_CVAR(Bool, render_things)
EXTERN_CVAR(Bool, render_fullbright)

// load_binds: Loads the key configuration file
// ----------------------------------------- >>
void load_binds()
{
	/*
	if (!binds.load("config/keys.cfg"))
	{
		message_box("Cannot find config/keys.cfg!");
		exit(0);
	}
	*/
	keys_defaults();

	if (!binds.load("config/keys.cfg"))
		printf("config/keys.cfg doesn't exist, setting default controls...\n\n");
}

// increase_grid: Increases the grid size
// ----------------------------------- >>
void increase_grid()
{
	gridsize *= 2;

	if (gridsize > 1024)
		gridsize = 1024;

	update_grid();
}

// decrease_grid: Decreases the grid size
// ----------------------------------- >>
void decrease_grid()
{
	gridsize /= 2;

	if (gridsize < 1)
		gridsize = 1;

	update_grid();
}

// edit_item: When the edit item key is pressed
// ----------------------------------------- >>
void edit_item()
{
	if (edit_mode == 0)
	{
		create_vertex();
		return;
	}

	if (hilight_item == -1 && !selection())
		return;

	if (edit_mode == 1)
	{
		open_menu = MENU_LINEMAIN;
		setup_line_menu();
		max_item = line_options.n_strings - 1;
	}

	if (edit_mode == 2)
	{
		open_menu = MENU_SECTORMAIN;
		max_item = sector_options.n_strings - 1;
	}

	if (edit_mode == 3)
	{
		open_menu = MENU_THINGMAIN;
		max_item = thing_options.n_strings - 1;
	}

	menu = true;
	sel_item = 0;
}

void change_edit_mode()
{
	if (++edit_mode > 3)
		edit_mode = 0;
}

// keys_edit: Keys for the 2d editor
// ------------------------------ >>
void keys_edit()
{
	keys = SDL_GetKeyState(NULL);
	
	// Scroll up
	if (binds.pressed("view_up", keys))
	{
		yoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Scroll down
	if (binds.pressed("view_down", keys))
	{
		yoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		update_grid();
		update_map();
	}

	// Scroll left
	if (binds.pressed("view_left", keys))
	{
		xoff += ((MAJOR_UNIT / (int)zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Scroll right
	if (binds.pressed("view_right", keys))
	{
		xoff -= ((MAJOR_UNIT / (int)zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Zoom in
	if (binds.pressed("view_zoomin", keys))
	{
		zoom *= 1.2f;

		if (zoom > 1024)
			zoom = 1024;

		update_grid();
		update_map();
	}
	
	// Zoom out
	if (binds.pressed("view_zoomout", keys))
	{
		zoom *= 0.8f;

		if (zoom < 1)
			zoom = 1;

		update_grid();
		update_map();
	}
	
	// Center view on mouse
	if (binds.pressed("view_mousecenter", keys))
	{
		xoff = -m_x(mouse.x) / MAJOR_UNIT;
		yoff = -m_y(mouse.y) / MAJOR_UNIT;
		update_grid();
		update_map();
	}

	// Set offsets to 0, 0
	if (binds.pressed("view_origin", keys))
	{
		xoff = yoff = 0;
		update_grid();
		update_map();
	}

	// Vertices mode
	if (binds.pressed("mode_vertices", keys))
	{
		clear_selection();
		edit_mode = 0;
		update_map();
	}
	
	// Linedefs mode
	if (binds.pressed("mode_linedefs", keys))
	{
		clear_selection();
		edit_mode = 1;
		update_map();
	}
	
	// Sectors mode
	if (binds.pressed("mode_sectors", keys))
	{
		clear_selection();
		edit_mode = 2;
		update_map();
	}
	
	// Things mode
	if (binds.pressed("mode_things", keys))
	{
		clear_selection();
		edit_mode = 3;
		update_map();
	}

	// Change mode
	if (binds.pressed("mode_change", keys))
	{
		clear_selection();
		change_edit_mode();
		update_map();
	}
 
	// Increase grid size
	if (binds.pressed("view_increasegrid", keys))
		increase_grid();
	
	// Decrease grid size
	if (binds.pressed("view_decreasegrid", keys))
		decrease_grid();
	
	// Clear selection
	if (binds.pressed("edit_clearselection", keys))
		clear_selection();

	/* Screenshot (crashes in GL for some reason)
	if (keys[SDLK_F12])
	{
		SDL_SaveBMP(screen, "shot.bmp");
		keys[SDLK_F12] = false;
	}
	*/

	if (binds.pressed("view_3dmode", keys) && map.n_sectors > 1)
	{
		keys[binds.key("view_3dmode")] = false;
		mode_3d = true;

		if (map_changed >= 2)
			cleanup_3dmode();

		if (map_changed == 3)
		{
			gui_drawsplash("Building GL nodes...");
			build_nodes();
		}

		if (map_changed >= 2)
		{
			gui_drawsplash("Setting up 3d mode...");
			setup_3dmode();
		}
		
		map_changed = 1;
		init_opengl();
	}

	if (binds.pressed("edit_edititem", keys))
	{
		edit_item();
	}

	if (keys[binds.key("edit_createitem")])
	{
		if (edit_mode == 0)
		{
			if (!selection())
				create_vertex();
			else
			{
				if (SDL_GetModState() & KMOD_SHIFT)
					create_lines(true);
				else
					create_lines(false);
			}

			return;
		}

		if (edit_mode == 1)
		{
			if (selection())
				create_sector();

			return;
		}

		if (edit_mode == 3)
		{
			create_thing();

			return;
		}

		keys[binds.key("edit_createitem")] = false;
	}

	if (binds.pressed("edit_undo", keys))
	{
		undo();
		update_map();
		map_changed = 3;
		keys[binds.key("edit_undo")] = false;
	}

	if (binds.pressed("line_begindraw", keys))
	{
		if (!line_draw)
			line_draw = true;

		keys[binds.key("line_begindraw")] = false;
	}

	if (binds.pressed("line_begindraw_rect", keys))
	{
		if (!line_draw)
		{
			line_draw = true;
			sel_box.set(mouse.x, mouse.y, mouse.x, mouse.y);
		}

		keys[binds.key("line_begindraw_rect")] = false;
	}

	if (binds.pressed("edit_deleteitem", keys))
	{
		if (edit_mode == 0)
			delete_vertex();

		if (edit_mode == 1)
			delete_line();

		if (edit_mode == 2)
			delete_sector();

		if (edit_mode == 3)
			delete_thing();
	}

	if (binds.pressed("line_flip", keys))
	{
		if (edit_mode == 1)
			line_flip(true, false);
	}

	if (binds.pressed("line_swapsides", keys))
	{
		if (edit_mode == 1)
			line_flip(false, true);
	}

	if (binds.pressed("line_flipboth", keys))
	{
		if (edit_mode == 1)
			line_flip(true, true);
	}

	// Sector height quick changes (8 units)
	if (binds.pressed("sector_upfloor8", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(true, 8);
	}

	if (binds.pressed("sector_downfloor8", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(true, -8);
	}

	if (binds.pressed("sector_upceil8", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(false, 8);
	}

	if (binds.pressed("sector_downceil8", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(false, -8);
	}

	if (binds.pressed("sector_upboth8", keys))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, 8);
			sector_changeheight(false, 8);
		}
	}

	if (binds.pressed("sector_downboth8", keys))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, -8);
			sector_changeheight(false, -8);
		}
	}

	// Sector height quick changes (1 unit)
	if (binds.pressed("sector_upfloor", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(true, 1);
	}

	if (binds.pressed("sector_downfloor", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(true, -1);
	}

	if (binds.pressed("sector_upceil", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(false, 1);
	}

	if (binds.pressed("sector_downceil", keys))
	{
		if (edit_mode == 2)
			sector_changeheight(false, -1);
	}

	if (binds.pressed("sector_upboth", keys))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, 1);
			sector_changeheight(false, 1);
		}
	}

	if (binds.pressed("sector_downboth", keys))
	{
		if (edit_mode == 2)
		{
			sector_changeheight(true, -1);
			sector_changeheight(false, -1);
		}
	}

	// Sector merge and join
	if (binds.pressed("sector_join", keys))
	{
		sector_merge(true);
		keys[binds.key("sector_join")] = false;
	}

	if (binds.pressed("sector_merge", keys))
	{
		sector_merge(false);
		keys[binds.key("sector_merge")] = false;
	}

	// Menu
	if (binds.pressed("menu_file", keys))
	{
		menu = true;
		open_menu = MENU_FILEMENU;
		sel_item = 0;
		max_item = 2;
		keys[binds.key("menu_file")] = false;
	}

	if (binds.pressed("menu_edit", keys))
	{
		menu = true;
		open_menu = MENU_EDITMENU;
		setup_edit_menu();
		sel_item = 0;
		max_item = edit_menu.n_strings - 1;
		keys[binds.key("menu_edit")] = false;
	}

	if (binds.pressed("menu_options", keys))
	{
		menu = true;
		open_menu = MENU_OPTIONSMENU;
		sel_item = 0;
		max_item = 4;
		keys[binds.key("menu_options")] = false;
	}

	if (binds.pressed("menu_check", keys))
	{
		menu = true;
		open_menu = MENU_CHECKMENU;
		sel_item = 0;
		max_item = 5;
		keys[binds.key("menu_check")] = false;
	}

	if (binds.pressed("edit_lockhilight", keys))
	{
		lock_hilight = !lock_hilight;
		keys[binds.key("edit_lockhilight")] = false;
	}

	/*
	if (binds.pressed("open_console", keys))
	{
		console = true;

		if (console_height <= 0.0f)
			console_height = 0.02f;

		keys[binds.key("open_console")] = false;
	}
	*/
}

// keys_linedraw: Keys for line drawing
// --------------------------------- >>
void keys_linedraw()
{
	keys = SDL_GetKeyState(NULL);

	// Scroll up
	if (binds.pressed("view_up", keys))
	{
		yoff += ((MAJOR_UNIT / zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Scroll down
	if (binds.pressed("view_down", keys))
	{
		yoff -= ((MAJOR_UNIT / zoom)) + 1;
		update_grid();
		update_map();
	}

	// Scroll left
	if (binds.pressed("view_left", keys))
	{
		xoff += ((MAJOR_UNIT / zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Scroll right
	if (binds.pressed("view_right", keys))
	{
		xoff -= ((MAJOR_UNIT / zoom)) + 1;
		update_grid();
		update_map();
	}
	
	// Zoom in
	if (binds.pressed("view_zoomin", keys))
	{
		zoom *= 2;
		
		if (zoom > 128)
			zoom = 128;

		update_grid();
		update_map();
	}
	
	// Zoom out
	if (binds.pressed("view_zoomout", keys))
	{
		zoom /= 2;

		if (zoom < 1)
			zoom = 1;

		update_grid();
		update_map();
	}
	
	// Center view on mouse
	if (binds.pressed("view_mousecenter", keys))
	{
		xoff = -m_x(mouse.x) / MAJOR_UNIT;
		yoff = -m_y(mouse.y) / MAJOR_UNIT;
		update_grid();
		update_map();
	}

	// Set offsets to 0, 0
	if (binds.pressed("view_origin", keys))
	{
		xoff = yoff = 0;
		update_grid();
		update_map();
	}

	// Increase grid size
	if (binds.pressed("view_increasegrid", keys))
		increase_grid();
	
	// Decrease grid size
	if (binds.pressed("view_decreasegrid", keys))
		decrease_grid();

	if (binds.pressed("edit_align_tex_x", keys))

	if (keys[SDLK_ESCAPE])
	{
		line_draw = false;
		end_linedraw();
	}
}

// keys_3d: Keys for 3d mode
// ---------------------- >>
void keys_3d()
{
	keys = SDL_GetKeyState(NULL);

	if (binds.pressed("3d_forward", keys))
	{
		camera.MoveCamera(move_speed_3d);
		determine_3d_hilight();

		//if (gravity)
			//camera_to_floor();
	}
	
	if (binds.pressed("3d_back", keys))
	{
		camera.MoveCamera(-move_speed_3d);
		determine_3d_hilight();

		//if (gravity)
			//camera_to_floor();
	}

	if (binds.pressed("3d_left", keys))
	{
		camera.RotateView(0.01f, 0.0f, 0.0f, 1.0f);
		determine_3d_hilight();
	}

	if (binds.pressed("3d_right", keys))
	{
		camera.RotateView(-0.01f, 0.0f, 0.0f, 1.0f);
		determine_3d_hilight();
	}

	if (binds.pressed("3d_strafeleft", keys))
	{
		camera.StrafeCamera(-move_speed_3d);
		determine_3d_hilight();

		//if (gravity)
			//camera_to_floor();
	}

	if (binds.pressed("3d_straferight", keys))
	{
		camera.StrafeCamera(move_speed_3d);
		determine_3d_hilight();

		//if (gravity)
			//camera_to_floor();
	}

	if (binds.pressed("3d_moveup", keys))
	{
		camera.position.z += move_speed_3d;
		camera.view.z += move_speed_3d;
		determine_3d_hilight();
	}

	if (binds.pressed("3d_movedown", keys))
	{
		camera.position.z -= move_speed_3d;
		camera.view.z -= move_speed_3d;
		determine_3d_hilight();
	}

	if (binds.pressed("3d_togglegravity", keys) && key_3d_rep == 0)
	{
		gravity = !gravity;

		msg_count = 100;
		msg_text = "Gravity";

		if (gravity)
		{
			msg_text += " On";
			//camera_to_floor();
		}
		else
			msg_text += " Off";

		key_3d_rep = KEY_3D_DELAY;
		determine_3d_hilight();
	}

	if (binds.pressed("open_console", keys))
	{
		console = true;

		if (console_height <= 0.0f)
			console_height = 0.02f;

		keys[binds.key("open_console")] = false;
	}

	if (binds.pressed("3d_2dmode", keys))
	{
		printf("Exit 3d mode...\n");
		mode_3d = false;
		init_opengl();
		draw_2d_screen();
	}

	// Change thing z-height
	if (binds.pressed("3d_upthingz8", keys) && key_3d_rep == 0)
	{
		if (change_3d_thing_z(8))
			key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downthingz8", keys) && key_3d_rep == 0)
	{
		if (change_3d_thing_z(-8))
			key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upthingz", keys) && key_3d_rep == 0)
	{
		if (change_3d_thing_z(1))
			key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downthingz", keys) && key_3d_rep == 0)
	{
		if (change_3d_thing_z(-1))
			key_3d_rep = KEY_3D_DELAY;
	}

	// Sector height quick changes (8 units)
	if (binds.pressed("3d_upfloor8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, 8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downfloor8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, -8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upceil8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(false, 8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downceil8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(false, -8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upboth8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, 8);
		sector_3d_changeheight(false, 8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downboth8", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, -8);
		sector_3d_changeheight(false, -8);
		key_3d_rep = KEY_3D_DELAY;
	}

	// Sector height quick changes (1 unit)
	if (binds.pressed("3d_upfloor", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, 1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downfloor", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, -1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upceil", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(false, 1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downceil", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(false, -1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upboth", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, 1);
		sector_3d_changeheight(false, 1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downboth", keys) && key_3d_rep == 0)
	{
		sector_3d_changeheight(true, -1);
		sector_3d_changeheight(false, -1);
		key_3d_rep = KEY_3D_DELAY;
	}

	// Light level
	if (binds.pressed("3d_upbrightness", keys) && key_3d_rep == 0)
	{
		change_3d_sector_light(16);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downbrightness", keys) && key_3d_rep == 0)
	{
		change_3d_sector_light(-16);
		key_3d_rep = KEY_3D_DELAY;
	}

	// Texture offset
	if (binds.pressed("3d_upyoffset", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(0, 1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downyoffset", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(0, -1);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upxoffset", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(1, 0);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downxoffset", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(-1, 0);
		key_3d_rep = KEY_3D_DELAY;
	}

	// Texture offset x8
	if (binds.pressed("3d_upyoffset8", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(0, 8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downyoffset8", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(0, -8);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_upxoffset8", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(8, 0);
		key_3d_rep = KEY_3D_DELAY;
	}

	if (binds.pressed("3d_downxoffset8", keys) && key_3d_rep == 0)
	{
		change_3d_texture_offset(-8, 0);
		key_3d_rep = KEY_3D_DELAY;
	}

	// Change thing angle
	if (binds.pressed("3d_upthingangle", keys))
	{
		change_3d_thingangle(45);
		keys[binds.key("3d_upthingangle")] = false;
	}

	if (binds.pressed("3d_downthingangle", keys))
	{
		change_3d_thingangle(-45);
		keys[binds.key("3d_downthingangle")] = false;
	}

	// Peg/Unpeg textures
	if (binds.pressed("3d_upperunpegged", keys))
	{
		toggle_3d_texpeg(true);
		keys[binds.key("3d_upperunpegged")] = false;
	}

	if (binds.pressed("3d_lowerunpegged", keys))
	{
		toggle_3d_texpeg(false);
		keys[binds.key("3d_lowerunpegged")] = false;
	}

	// Clear selected wall texture
	if (binds.pressed("3d_clear_texture", keys))
	{
		delete_3d_texture();
		keys[binds.key("3d_clear_texture")] = false;
	}

	// Clear selected wall texture offsets
	if (binds.pressed("3d_reset_offsets", keys))
	{
		reset_3d_tex_offsets();
		keys[binds.key("3d_reset_offsets")] = false;
	}

	// Align textures on x starting at hilighted line
	if (binds.pressed("3d_align_tex_x", keys))
	{
		align_3d_tex();
		keys[binds.key("3d_align_tex_x")] = false;
	}

	// Toggle thing rendering
	if (binds.pressed("3d_toggle_things", keys))
	{
		render_things = !render_things;
		keys[binds.key("3d_toggle_things")] = false;
	}

	// Toggle fullbright mode
	if (binds.pressed("3d_toggle_fullbright", keys))
	{
		render_fullbright = !render_fullbright;
		keys[binds.key("3d_toggle_fullbright")] = false;
	}

	key_3d_rep--;
	if (key_3d_rep < 0)
		key_3d_rep = 0;
}

// keys_menu: Keys for menus
// ---------------------- >>
bool keys_menu()
{
	keys = SDL_GetKeyState(NULL);

	if (keys[SDLK_UP])
	{
		if (sel_item == 0)
			sel_item = max_item;
		else
			sel_item--;

		keys[SDLK_UP] = false;
	}

	if (keys[SDLK_DOWN])
	{
		if (sel_item == max_item)
			sel_item = 0;
		else
			sel_item++;

		keys[SDLK_DOWN] = false;
	}

	if (keys[SDLK_RETURN])
	{
		keys[SDLK_RETURN] = false;
		if (!menu_action())
			return false;

		update_map();
	}

	if (keys[SDLK_ESCAPE])
		menu = false;

	return true;
}

// keys_browser: Keys for the texture browser
// --------------------------------------- >>
bool keys_browser()
{
	keys = SDL_GetKeyState(NULL);

	if (keys[SDLK_ESCAPE])
		return false;

	if (keys[SDLK_LEFT])
		browser_changetex(0);

	if (keys[SDLK_RIGHT])
		browser_changetex(1);

	if (keys[SDLK_UP])
		browser_changetex(2);

	if (keys[SDLK_DOWN])
		browser_changetex(3);

	// 0-9
	for (char c = 48; c < 58; c++)
	{
		if (keys[c] && browser_search.length() < 8)
		{
			browser_search += c;
			browser_findtex(browser_search);
			keys[c] = false;
			//break;
		}
	}

	// A-Z (or a-z)
	for (char c = 97; c < 122; c++)
	{
		if (keys[c] && browser_search.length() < 8)
		{
			browser_search += c - 32;
			browser_findtex(browser_search);
			keys[c] = false;
			//break;
		}
	}

	// '-' or '_'
	if (keys[45])
	{
		if (keys[SDLK_LSHIFT] || keys[SDLK_RSHIFT])
			browser_search += '_';
		else
			browser_search += '-';

		browser_findtex(browser_search);
		keys[45] = false;
	}

	if (keys[SDLK_BACKSPACE])
	{
		if (browser_search.length() > 0)
			browser_search.erase(browser_search.length() - 1, 1);
	}

	if (browser_search.length() > 8)
		browser_search.resize(8);

	if (keys[SDLK_RETURN])
	{
		if (mode_3d)
			set_3d_texture(SDL_GetModState() & KMOD_SHIFT);
		else
			menu_action();

		return false;
	}

	return true;
}

// keys_inputbox: Keys for the input box
// ---------------------------------- >>
void keys_inputbox()
{
	keys = SDL_GetKeyState(NULL);

	// 0-9
	for (char c = 48; c < 58; c++)
	{
		if (keys[c])
		{
			if (!inbox_typed)
			{
				menu_input = "";
				inbox_typed = true;
			}

			menu_input += c;
			keys[c] = false;
			break;
		}
	}

	// A-Z (or a-z)
	for (char c = 97; c < 123; c++)
	{
		if (keys[c])
		{
			if (!inbox_typed)
			{
				menu_input = "";
				inbox_typed = true;
			}

			if (keys[SDLK_LSHIFT] || keys[SDLK_RSHIFT])
				menu_input += c - 32;
			else
				menu_input += c;
			
			keys[c] = false;
			break;
		}
	}

	if (keys[SDLK_RETURN])
	{
		inbox = false;
		inbox_typed = false;
		menu_inbox_action();
	}

	if (keys[SDLK_ESCAPE])
	{
		inbox = false;
		inbox_typed = false;
	}

	if (keys[SDLK_BACKSPACE])
	{
		if (!inbox_typed)
		{
			menu_input = "";
			inbox_typed = true;
		}

		if (menu_input.length() > 0)
			menu_input.erase(menu_input.length() - 1, 1);
	}

	// '-'
	if (keys[45])
	{
		if (!inbox_typed)
		{
			menu_input = "";
			inbox_typed = true;
		}

		menu_input += '-';
		keys[45] = false;
	}

	// '.'
	if (keys['.'])
	{
		if (!inbox_typed)
		{
			menu_input = "";
			inbox_typed = true;
		}

		menu_input += '.';
		keys['.'] = false;
	}

	if (menu_input.length() > 16)
		menu_input.resize(16);
}

// keys_console: Keys for the console
// ------------------------------- >>
void keys_console()
{
	keys = SDL_GetKeyState(NULL);

	// 0-9
	for (char c = 48; c < 58; c++)
	{
		if (keys[c])
		{
			console_command += c;
			keys[c] = false;
			//break;
		}
	}

	// A-Z (or a-z)
	for (char c = 97; c < 123; c++)
	{
		if (keys[c])
		{
			if (keys[SDLK_LSHIFT] || keys[SDLK_RSHIFT])
				console_command += c - 32;
			else
				console_command += c;
			
			keys[c] = false;
			//break;
		}
	}

	// '-' or '_'
	if (keys[45])
	{
		if (keys[SDLK_LSHIFT] || keys[SDLK_RSHIFT])
			console_command += '_';
		else
			console_command += '-';

		keys[45] = false;
	}

	// '.'
	if (keys['.'])
	{
		console_command += '.';
		keys['.'] = false;
	}

	// '"'
	if (keys['\''])
	{
		if (keys[SDLK_LSHIFT] || keys[SDLK_RSHIFT])
			console_command += "\"";

		keys['\''] = false;
	}

	// Space
	if (keys[SDLK_SPACE])
	{
		console_command += ' ';
		keys[SDLK_SPACE] = false;
	}

	if (keys[SDLK_UP])
	{
		console_prevcommand();
		keys[SDLK_UP] = false;
	}

	if (keys[SDLK_DOWN])
	{
		console_nextcommand();
		keys[SDLK_DOWN] = false;
	}

	if (keys[SDLK_RETURN])
	{
		console_parsecommand();
	}

	if (keys[SDLK_BACKSPACE])
	{
		if (console_command.length() > 0)
			console_command.erase(console_command.length() - 1, 1);
	}

	if (binds.pressed("open_console", keys))
	{
		console = false;
		keys[binds.key("open_console")] = false;
	}

	if (console_command.length() > 256)
		console_command.resize(256);
}

