// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004                         >>
// << ------------------------------------------- >>
// << <keybind.cpp>                               >>
// << KeyBind class functions, handles keys.cfg   >>
// <<--------------------------------------------->>

// Includes ------------------------------------- >>
#include "main.h"
#include "keybind.h"

// Structs -------------------------------------- >>
struct key_identifier_t
{
	string name;
	Uint16 key;

	key_identifier_t() { name = ""; key = 0; }
	key_identifier_t(string name, Uint16 key)
	{
		this->name = name;
		this->key = key;
	}
};

// Variables ------------------------------------ >>
BindList	binds;

// Key Identifiers
const BYTE NKEYID = 41;
key_identifier_t key_ids[NKEYID] = {
	key_identifier_t("KEY_UP",			273),
	key_identifier_t("KEY_DOWN",		274),
	key_identifier_t("KEY_LEFT",		276),
	key_identifier_t("KEY_RIGHT",		275),
	key_identifier_t("KEY_RETURN",		13),
	key_identifier_t("KEY_INSERT",		277),
	key_identifier_t("KEY_DELETE",		127),
	key_identifier_t("KEY_PGUP",		280),
	key_identifier_t("KEY_PGDN",		281),
	key_identifier_t("KEY_END",			279),
	key_identifier_t("KEY_HOME",		278),
	key_identifier_t("KEY_SPACE",		32),
	key_identifier_t("KEY_ESCAPE",		27),
	key_identifier_t("KEY_TAB",			9),
	key_identifier_t("KEYPAD_UP",		264),
	key_identifier_t("KEYPAD_DOWN",		258),
	key_identifier_t("KEYPAD_LEFT",		260),
	key_identifier_t("KEYPAD_RIGHT",	262),
	key_identifier_t("KEYPAD_PGUP",		265),
	key_identifier_t("KEYPAD_PGDN",		259),
	key_identifier_t("KEYPAD_HOME",		263),
	key_identifier_t("KEYPAD_END",		257),
	key_identifier_t("KEYPAD_PLUS",		270),
	key_identifier_t("KEYPAD_MINUS",	269),
	key_identifier_t("KEYPAD_/",		267),
	key_identifier_t("KEYPAD_*",		268),
	key_identifier_t("KEYPAD_ENTER",	271),
	key_identifier_t("KEYPAD_DEL",		266),
	key_identifier_t("KEYPAD_INS",		256),
	key_identifier_t("KEY_F1",			282),
	key_identifier_t("KEY_F2",			283),
	key_identifier_t("KEY_F3",			284),
	key_identifier_t("KEY_F4",			285),
	key_identifier_t("KEY_F5",			286),
	key_identifier_t("KEY_F6",			287),
	key_identifier_t("KEY_F7",			288),
	key_identifier_t("KEY_F8",			289),
	key_identifier_t("KEY_F9",			290),
	key_identifier_t("KEY_F10",			291),
	key_identifier_t("KEY_F11",			292),
	key_identifier_t("KEY_F12",			293)
};

// key_id: Gets the keycode for a key id name
// --------------------------------------- >>
Uint16 key_id(string name)
{
	for (BYTE a = 0; a < NKEYID; a++)
	{
		if (key_ids[a].name == name)
			return key_ids[a].key;
	}

	return 0;
}

// keys_defaults: Sets all key binds to defaults
// ------------------------------------------ >>
void keys_defaults()
{
	binds.add("menu_file", 'f', KEYMOD_ALT);
	binds.add("menu_edit", 'e', KEYMOD_ALT);
	binds.add("menu_options", 'o', KEYMOD_ALT);
	binds.add("menu_mode", 'm', KEYMOD_ALT);
	binds.add("menu_check", 'c', KEYMOD_ALT);

	binds.add("view_up", key_id("KEY_UP"), 0);
	binds.add("view_down", key_id("KEY_DOWN"), 0);
	binds.add("view_left", key_id("KEY_LEFT"), 0);
	binds.add("view_right", key_id("KEY_RIGHT"), 0);
	binds.add("view_zoomin", '=', 0);
	binds.add("view_zoomout", '-', 0);
	binds.add("view_origin", 'o', 0);
	binds.add("view_increasegrid", '[', 0);
	binds.add("view_decreasegrid", ']', 0);
	binds.add("view_3dmode", 'q', 0);

	binds.add("mode_vertices", 'v', 0);
	binds.add("mode_linedefs", 'l', 0);
	binds.add("mode_sectors", 's', 0);
	binds.add("mode_things", 't', 0);
	binds.add("mode_change", key_id("KEY_TAB"), 0);

	binds.add("edit_lockhilight", 'h', 0);
	binds.add("edit_clearselection", 'c', 0);
	binds.add("edit_edititem", key_id("KEY_RETURN"), 0);
	binds.add("edit_createitem", key_id("KEY_INSERT"), 0);
	binds.add("edit_deleteitem", key_id("KEY_DELETE"), 0);
	binds.add("edit_undo", 'z', KEYMOD_CTRL);

	binds.add("line_flip", 'f', 0);
	binds.add("line_swapsides", 'f', KEYMOD_CTRL);
	binds.add("line_flipboth", 'f', KEYMOD_CTRL|KEYMOD_SHIFT);
	binds.add("line_begindraw", key_id("KEY_SPACE"), 0);
	binds.add("line_begindraw_rect", key_id("KEY_SPACE"), KEYMOD_SHIFT);

	binds.add("sector_upfloor8", '.', 0);
	binds.add("sector_downfloor8", ',', 0);
	binds.add("sector_upceil8", '.', KEYMOD_SHIFT);
	binds.add("sector_downceil8", ',', KEYMOD_SHIFT);
	binds.add("sector_upfloor", '.', KEYMOD_CTRL);
	binds.add("sector_downfloor", ',', KEYMOD_CTRL);
	binds.add("sector_upceil", '.', KEYMOD_CTRL|KEYMOD_SHIFT);
	binds.add("sector_downceil", ',', KEYMOD_CTRL|KEYMOD_SHIFT);
	binds.add("sector_upboth8", '.', KEYMOD_ALT);
	binds.add("sector_downboth8", ',', KEYMOD_ALT);
	binds.add("sector_upboth", '.', KEYMOD_ALT|KEYMOD_CTRL);
	binds.add("sector_downboth", ',', KEYMOD_ALT|KEYMOD_CTRL);
	binds.add("sector_merge", 'j', KEYMOD_CTRL);
	binds.add("sector_join", 'j', 0);

	binds.add("3d_forward", 'w', 0);
	binds.add("3d_back", 's', 0);
	binds.add("3d_strafeleft", 'a', 0);
	binds.add("3d_straferight", 'd', 0);
	binds.add("3d_left", key_id("KEY_LEFT"), 0);
	binds.add("3d_right", key_id("KEY_RIGHT"), 0);
	binds.add("3d_moveup", key_id("KEY_UP"), 0);
	binds.add("3d_movedown", key_id("KEY_DOWN"), 0);
	binds.add("3d_togglegravity", 'g', 0);

	binds.add("3d_2dmode", 'q', 0);
	binds.add("3d_toggle_things", 't', 0);
	binds.add("3d_toggle_fullbright", 'b', 0);

	binds.add("3d_upceil8", key_id("KEYPAD_PLUS"), 0);
	binds.add("3d_downceil8", key_id("KEYPAD_MINUS"), 0);
	binds.add("3d_upceil", key_id("KEYPAD_PLUS"), KEYMOD_CTRL);
	binds.add("3d_downceil", key_id("KEYPAD_MINUS"), KEYMOD_CTRL);
	binds.add("3d_upfloor8", key_id("KEYPAD_PLUS"), KEYMOD_SHIFT);
	binds.add("3d_downfloor8", key_id("KEYPAD_MINUS"), KEYMOD_SHIFT);
	binds.add("3d_upfloor", key_id("KEYPAD_PLUS"), KEYMOD_SHIFT|KEYMOD_CTRL);
	binds.add("3d_downfloor", key_id("KEYPAD_MINUS"), KEYMOD_SHIFT|KEYMOD_CTRL);
	binds.add("3d_upboth8", key_id("KEYPAD_PLUS"), KEYMOD_ALT);
	binds.add("3d_downboth8", key_id("KEYPAD_MINUS"), KEYMOD_ALT);
	binds.add("3d_upboth", key_id("KEYPAD_PLUS"), KEYMOD_ALT|KEYMOD_CTRL);
	binds.add("3d_downboth", key_id("KEYPAD_MINUS"), KEYMOD_ALT|KEYMOD_CTRL);

	binds.add("3d_upbrightness", '=', 0);
	binds.add("3d_downbrightness", '-', 0);

	binds.add("3d_upyoffset", key_id("KEYPAD_UP"), KEYMOD_CTRL);
	binds.add("3d_downyoffset", key_id("KEYPAD_DOWN"), KEYMOD_CTRL);
	binds.add("3d_upxoffset", key_id("KEYPAD_LEFT"), KEYMOD_CTRL);
	binds.add("3d_downxoffset", key_id("KEYPAD_RIGHT"), KEYMOD_CTRL);
	binds.add("3d_upyoffset8", key_id("KEYPAD_UP"), 0);
	binds.add("3d_downyoffset8", key_id("KEYPAD_DOWN"), 0);
	binds.add("3d_upxoffset8", key_id("KEYPAD_LEFT"), 0);
	binds.add("3d_downxoffset8", key_id("KEYPAD_RIGHT"), 0);

	binds.add("3d_upthingangle", key_id("KEYPAD_LEFT"), 0);
	binds.add("3d_downthingangle", key_id("KEYPAD_RIGHT"), 0);

	binds.add("3d_upthingz8", key_id("KEYPAD_UP"), 0);
	binds.add("3d_downthingz8", key_id("KEYPAD_DOWN"), 0);
	binds.add("3d_upthingz", key_id("KEYPAD_UP"), KEYMOD_CTRL);
	binds.add("3d_downthingz", key_id("KEYPAD_DOWN"), KEYMOD_CTRL);

	binds.add("3d_lowerunpegged", 'l', 0);
	binds.add("3d_upperunpegged", 'u', 0);
	binds.add("3d_clear_texture", key_id("KEY_DELETE"), 0);
	binds.add("3d_reset_offsets", 'r', 0);
	binds.add("3d_align_tex_x", 'a', KEYMOD_CTRL);

	binds.add("open_console", '`', 0);
}

// BindList::add: Adds a key to the binds list
// ---------------------------------------- >>
void BindList::add(string name, Uint16 key, Uint8 mod)
{
	keybind_t	*temp = first;
	
	first = new keybind_t;

	first->key = key;
	first->name = name;
	first->mod = mod;
	first->next = temp;
}

// BindList::add: Adds a key to the binds list
// ---------------------------------------- >>
void BindList::set(string name, Uint16 key, Uint8 mod)
{
	keybind_t	*temp = first;
	
	while (temp)
	{
		if (temp->name == name)
		{
			temp->key = key;
			temp->mod = mod;
			return;
		}
		
		temp = temp->next;
	}

	return;
}

// BindList::key: Gets the key code for a command
// ------------------------------------------- >>
Uint16 BindList::key(string name)
{
	keybind_t	*temp = first;
	
	while (temp)
	{
		if (temp->name == name)
			return temp->key;
		
		temp = temp->next;
	}
	
	return 1;
}

// BindList::pressed: Determines if the key for a command is pressed
// (including any key modifiers shift/ctrl/alt)
// -------------------------------------------------------------- >>
bool BindList::pressed(string name, Uint8* keys)
{
	keybind_t	*temp = first;

	while (temp)
	{
		if (temp->name == name)
			break;

		temp = temp->next;
	}

	if (!temp)
		return false;

	if (keys[temp->key])
	{
		bool rightkey = true;

		if (temp->mod & KEYMOD_SHIFT)
		{
			if (!(SDL_GetModState() & KMOD_SHIFT))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_SHIFT)
				rightkey = false;
		}

		if (temp->mod & KEYMOD_CTRL)
		{
			if (!(SDL_GetModState() & KMOD_CTRL))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_CTRL)
				rightkey = false;
		}

		if (temp->mod & KEYMOD_ALT)
		{
			if (!(SDL_GetModState() & KMOD_ALT))
				rightkey = false;
		}
		else
		{
			if (SDL_GetModState() & KMOD_ALT)
				rightkey = false;
		}

		return rightkey;
	}

	return false;
}

// BindList::load: Loads key binds from a file
// ---------------------------------------- >>
bool BindList::load(char* filename)
{
	MemReader mr;
	string token;
	string name;
	Uint8 mod;

	if (!mr.open_file(filename, 0, 0))
	{
		printf("keys.cfg not found!\n");
		return false;
	}

	token = mr.get_token();

	while (token != "!END")
	{
		mod = 0;
		name = token;

		// Read modifiers
		bool done = false;

		while (!done)
		{
			token = mr.get_token();

			if (token == "SHIFT")
				mod |= KEYMOD_SHIFT;
			else if (token == "CTRL")
				mod |= KEYMOD_CTRL;
			else if (token == "ALT")
				mod |= KEYMOD_ALT;
			else
				done = true;
		}

		// Add key
		Uint16 key = 0;

		if ((Uint16)token[0] > 47 && (Uint16)token[0] < 48)
			key = (Uint16)atoi(token.c_str());
		else
		{
			key = key_id(token);

			if (key == 0)
				key = (Uint16)token[0];
		}

		set(name, key, mod);
	
		// Apply modifiers
		//first->mod = mod;

		string mods;

		if (mod & KEYMOD_SHIFT)
			mods += "SHIFT";
		if (mod & KEYMOD_CTRL)
			mods += "CTRL";
		if (mod & KEYMOD_ALT)
			mods += "ALT";

		//printf("Added key %s = %d, with modifiers %s\n", name.c_str(), first->key, mods.c_str());

		token = mr.get_token();
	}

	return true;
}

string get_key_string(Uint16 keycode)
{
	if (keycode > 32 && keycode < 127)
		return string(1, (char)keycode);

	for (BYTE a = 0; a < NKEYID; a++)
	{
		if (key_ids[a].key == keycode)
			return key_ids[a].name;
	}

	char codestr[6] = "";
	sscanf(codestr, "%d", keycode);

	return codestr;
}

void BindList::save(char *filename)
{
	FILE *fp = fopen(filename, "wt");

	fprintf(fp, "// SLADE key configuration file\n");
	fprintf(fp, "// ----------------------------\n\n");
	fprintf(fp, "// The format isn't too hard to work out.\n");
	fprintf(fp, "// Valid special keys are:\n");
	
	fprintf(fp, "// ");
	for (BYTE a = 0; a < NKEYID; a++)
	{
		fprintf(fp, "%s", key_ids[a].name.c_str());

		if (a != NKEYID - 1)
		{
			if (a % 4 == 0)
				fprintf(fp, "\n// ");
			else
				fprintf(fp, ", ");
		}

	}
	fprintf(fp, "\n\n");

	fprintf(fp, "// You can also use numerical keycodes.\n\n");
	fprintf(fp, "// Also to specify modifier keys:\n");
	fprintf(fp, "// ALT CTRL SHIFT (before key)\n\n");

	keybind_t	*temp = first;

	while (temp)
	{
		string line = temp->name;
		line += " ";

		if (temp->mod & KEYMOD_SHIFT)
			line += "SHIFT ";
		if (temp->mod & KEYMOD_CTRL)
			line += "CTRL ";
		if (temp->mod & KEYMOD_ALT)
			line += "ALT ";

		fprintf(fp, "%s %s\n", line.c_str(), get_key_string(temp->key).c_str());

		temp = temp->next;
	}

	fclose(fp);
}

void BindList::dump_to_list(stringlist_t *list)
{
	keybind_t *temp = first;

	while (temp)
	{
		list->add(temp->name);
		temp = temp->next;
	}
}
