// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004-05                      >>
// << ------------------------------------------- >>
// << <menu.cpp>                                  >>
// << The horribly hack-y menu system             >>
// << (but it serves its purpose, so whatever ;)) >>
// <<--------------------------------------------->>

// Disable conversion warnings
#pragma warning (disable: 4244)

// Includes ------------------------------------- >>
#include "main.h"
#include "menu.h"
#include "gui.h"
#include "checks.h"
#include "map.h"
#include "tex_browse.h"
#include "textures.h"
#include "input.h"
#include "edit.h"
#include "keybind.h"

// Variables ------------------------------------ >>
BYTE	sel_item = 0;			// The currently selected menu item
BYTE	max_item = 0;			// The number of items in the current menu
BYTE	open_menu = 0;			// Which menu is open (see menu.h)
BYTE	thing_group=0;			// The currently selected thing type group
BYTE	line_group = 0;			// The currently selected line type group
BYTE	line_side = 1;			// Line side to edit (1=front, 2=back)
BYTE	sel_arg = 0;			// Selected line/thing arg for modification
bool	menu = false;			// Is a menu currently active?
bool	inbox = false;			// Is an input box showing?
bool	inbox_typed = false;	// Has the input box been typed in?
bool	message = false;		// Is a popup message showing?
bool	sel_items[16];			// State of items in a toggle menu
string	menu_input;				// String to hold input from inputboxes
rect_t	menu_area;				// The area taken up by the current menu

// StringLists for the menus
stringlist_t thing_options;
stringlist_t thing_angles;
stringlist_t thing_flags;
stringlist_t thing_groups;
stringlist_t thing_types[32];
stringlist_t thing_args;

stringlist_t line_options;
stringlist_t line_flags;
stringlist_t line_groups;
stringlist_t line_types[64];
stringlist_t line_args;
stringlist_t line_triggers;
stringlist_t line_arg_type;

stringlist_t side_options;

stringlist_t sector_options;
stringlist_t sector_typelist;

stringlist_t file_menu;
stringlist_t edit_menu;
stringlist_t options_menu;
stringlist_t mode_menu;
stringlist_t check_menu;

stringlist_t skin_menu;

stringlist_t display_menu;
stringlist_t video_menu;
stringlist_t filter_menu;

stringlist_t browser_menu;
stringlist_t browser_order_menu;

stringlist_t controls_menu;
stringlist_t controls_2dmode;
stringlist_t controls_3dmode;
stringlist_t controls_misc;

// Popup message
stringlist_t	popup_message;
string		message_title;

// Map objects
thing_t	last_thing;

// External Variables --------------------------- >>
extern Map map;
extern WadList wads;
extern int hilight_item, map_changed;
extern bool	browser, mix_tex;
extern int sel_trans;
extern stringlist_t colour_schemes, skin_dirs;
extern point2_t mouse;
extern int edit_mode;
extern numlist_t selected_items;
extern movelist_t move_list;
extern BindList binds;
//extern String texname;

EXTERN_CVAR(Bool, line_aa)
EXTERN_CVAR(Float, line_size)
EXTERN_CVAR(Bool, thing_sprites)
EXTERN_CVAR(Bool, grid_dashed)
EXTERN_CVAR(Int, vid_fullscreen)
EXTERN_CVAR(Int, vid_width)
EXTERN_CVAR(Int, vid_height)
EXTERN_CVAR(Int, vid_bitdepth)
EXTERN_CVAR(Int, gl_filter)
EXTERN_CVAR(String, ui_skin)
EXTERN_CVAR(Int, browser_columns)
EXTERN_CVAR(Int, browser_order)

// menu_pre_init: Init menu stuff (before loading line/thing types)
// ------------------------------------------------------------- >>
void menu_pre_init()
{
	line_groups.add("No Special (Normal)");
}

// menu_init: Initialise menu stuff
// ----------------------------- >>
void menu_init()
{
	thing_options.add("Change Thing Type");
	thing_options.add("Change Thing Angle");
	thing_options.add("Change Thing Flags");

	if (map.hexen)
	{
		thing_options.add("Change Thing ID");
		thing_options.add("Change Thing Special");
		thing_options.add("Change Thing Args");
		thing_options.add("Change Thing Z Height");
	}

	thing_options.add("Browse Thing Types");

	thing_groups.add("Enter Type");

	thing_angles.add("East");
	thing_angles.add("Northeast");
	thing_angles.add("North");
	thing_angles.add("Northwest");
	thing_angles.add("West");
	thing_angles.add("Southwest");
	thing_angles.add("South");
	thing_angles.add("Southeast");

	thing_flags.add("Appears On Easy Skill");
	thing_flags.add("Appears On Medium Skill");
	thing_flags.add("Appears On Hard Skill");
	thing_flags.add("Deaf (Only Wakes On Seeing)");

	if (!map.hexen)
		thing_flags.add("Multiplayer Only");
	else
	{
		thing_flags.add("Dormant");
		thing_flags.add("Appears To Fighter Class");
		thing_flags.add("Appears To Cleric Class");
		thing_flags.add("Appears To Mage Class");
		thing_flags.add("Appears In Single Player Mode");
		thing_flags.add("Appears In Cooperative Mode");
		thing_flags.add("Appears In Deathmatch Mode");

		if (map.zdoom)
		{
			thing_flags.add("Translucent");
			thing_flags.add("Invisible");
			thing_flags.add("Friendly");
			thing_flags.add("Stands Still");
		}
	}

	line_flags.add("Impassible");
	line_flags.add("Block Monsters");
	line_flags.add("Two Sided");
	line_flags.add("Upper Unpegged");
	line_flags.add("Lower Unpegged");
	line_flags.add("Secret Line");
	line_flags.add("Block Sound");
	line_flags.add("Don't Show On Map");
	line_flags.add("Start On Map");

	if (map.hexen)
		line_flags.add("Repeatable");

	if (map.zdoom)
	{
		line_flags.add("Monsters Can Activate");
		line_flags.add("Block Everything");
	}

	line_groups.add("Enter Type");

	line_triggers.add("Player Crosses Line");
	line_triggers.add("Player Uses Line");
	line_triggers.add("Monster Crosses Line");
	line_triggers.add("Projectile Hits Line");
	line_triggers.add("Player Bumps Line");
	line_triggers.add("Projectile Crosses Line");

	if (map.zdoom)
	{
		line_triggers.add("Player Uses Line (PassThru)");
		line_triggers.add("Projectile Hits OR Crosses Line");
	}

	side_options.add("Change Texture X Offset");
	side_options.add("Change Texture Y Offset");
	side_options.add("Change Middle Texture");
	side_options.add("Change Upper Texture");
	side_options.add("Change Lower Texture");
	side_options.add("Change Sector Reference");

	sector_options.add("Change Floor Height");
	sector_options.add("Change Ceiling Height");
	sector_options.add("Change Floor Texture");
	sector_options.add("Change Ceiling Texture");
	sector_options.add("Change Light Level");
	sector_options.add("Change Special");
	sector_options.add("Change Tag");

	file_menu.add("Save Map");
	file_menu.add("Save Map To New Wad...");
	file_menu.add("Close Map");

	options_menu.add("Display Options");
	options_menu.add("3d Mode Options");
	options_menu.add("Browser Options");
	options_menu.add("Set Video Mode");
	options_menu.add("Customize controls");

	browser_menu.add("Set Columns");
	browser_menu.add("Set Order");

	browser_order_menu.add("Alphabetical");
	browser_order_menu.add("Frequency Of Use");

	video_menu.add("800x600, 16bpp");
	video_menu.add("800x600, 32bpp");
	video_menu.add("1024x768, 16bpp");
	video_menu.add("1024x768, 32bpp");
	video_menu.add("1152x864, 16bpp");
	video_menu.add("1152x864, 32bpp");
	video_menu.add("1280x960, 16bpp");
	video_menu.add("1280x960, 32bpp");
	video_menu.add("1280x1024, 16bpp");
	video_menu.add("1280x1024, 32bpp");
	video_menu.add("1600x1200, 16bpp");
	video_menu.add("1600x1200, 32bpp");

	filter_menu.add("None (GL_NEAREST)");
	filter_menu.add("Linear (GL_LINEAR)");
	filter_menu.add("Trilinear (GL_LINEAR_MIPMAP_LINEAR)");

	mode_menu.add("Vertices");
	mode_menu.add("Linedefs");
	mode_menu.add("Sectors");
	mode_menu.add("Things");
	//mode_menu.add("3d Mode");

	check_menu.add("Map Statistics");
	check_menu.add("Clean Tags");
	check_menu.add("Clean Unused Vertices");
	check_menu.add("Clean Zero-Length Lines");
	check_menu.add("Clean Unused Sectors");
	check_menu.add("Check If All Sectors Are Closed");
	check_menu.add("Check Textures");

	controls_menu.add("2d Editing Controls");
	controls_menu.add("3d Editing Controls");
	controls_menu.add("Misc Controls");
}

// setup_control_menus: Adds key controls to appropriate control setup menus
// ---------------------------------------------------------------------- >>
void setup_control_menus()
{
	// Get all controls in a stringlist
	stringlist_t all_controls;
	binds.dump_to_list(&all_controls);

	// Populate control menus
	for (DWORD a = 0; a < all_controls.n_strings; a++)
	{
		// 2d mode keys
		if (all_controls.strings[a]->substr(0, 4) == "edit"
			|| all_controls.strings[a]->substr(0, 4) == "mode"
			|| all_controls.strings[a]->substr(0, 4) == "line"
			|| all_controls.strings[a]->substr(0, 4) == "view"
			|| all_controls.strings[a]->substr(0, 6) == "sector")
			controls_2dmode.add(*all_controls.strings[a]);
		// 3d mode keys
		else if (all_controls.strings[a]->substr(0, 2) == "3d")
			controls_3dmode.add(*all_controls.strings[a]);
		// Misc keys
		else
			controls_misc.add(*all_controls.strings[a]);
	}
}

// setup_edit_menu: Sets up the edit menu for the current editing mode
// ---------------------------------------------------------------- >>
void setup_edit_menu()
{
	edit_menu.clear();
	edit_menu.add("Edit Item");
	
	// Lines
	if (edit_mode == 1)
	{
		edit_menu.add("Extrude Lines");
	}

	// Sectors
	if (edit_mode == 2)
	{
		edit_menu.add("Create Stairs");
		edit_menu.add("Create Stairs (Floor + Ceil)");
		edit_menu.add("Create Door");
	}
}

// setup_display_menu: Sets up the display menu
// ----------------------------------------- >>
void setup_display_menu()
{
	display_menu.clear();
	display_menu.add("Choose Colour Scheme");
	display_menu.add("Choose Skin");
	display_menu.add("Line Antialiasing: " + bool2yesno(line_aa));
	display_menu.add("Set Line Width");
	display_menu.add("Dashed Grid: " + bool2yesno(grid_dashed));
	display_menu.add("Things as Sprites: " + bool2yesno(thing_sprites));
	display_menu.add("Set GL Filter");
	display_menu.add("Fullscreen: " + bool2yesno(vid_fullscreen));
}

// setup_line_menu: Sets up the main line menu depending on what sides the hilighted line has
// --------------------------------------------------------------------------------------- >>
void setup_line_menu()
{
	line_options.clear();

	if (hilight_item == -1)
	{
		line_options.add("Edit First Side");
		line_options.add("Edit Second Side");
	}
	else
	{
		if (map.lines[hilight_item]->side1 != -1)
			line_options.add("Edit First Side");
		else
			line_options.add("Add First Side");

		if (map.lines[hilight_item]->side2 != -1)
			line_options.add("Edit Second Side");
		else
			line_options.add("Add Second Side");
	}

	line_options.add("Change Sector Tag");
	line_options.add("Change Line Special");
	line_options.add("Change Line Flags");

	if (map.hexen)
	{
		line_options.add("Change Line Args");
		line_options.add("Change Line Trigger");
	}
}

// setup_line_args: Sets up the line args menu
// ---------------------------------------- >>
void setup_line_args()
{
	line_args.clear();

	if (hilight_item != -1)
	{
		line_args.add(get_ltype(map.lines[hilight_item]->type)->args[0]);
		line_args.add(get_ltype(map.lines[hilight_item]->type)->args[1]);
		line_args.add(get_ltype(map.lines[hilight_item]->type)->args[2]);
		line_args.add(get_ltype(map.lines[hilight_item]->type)->args[3]);
		line_args.add(get_ltype(map.lines[hilight_item]->type)->args[4]);
	}
	else
	{
		for (int a = 0; a < 5; a++)
			line_args.add("Arg" + a);
	}
}

// setup_line_args: Sets up the thing args menu
// ---------------------------------------- >>
void setup_thing_args()
{
	thing_args.clear();

	if (hilight_item != -1)
	{
		thing_args.add(get_tdef(map.things[hilight_item]->type)->args[0]);
		thing_args.add(get_tdef(map.things[hilight_item]->type)->args[1]);
		thing_args.add(get_tdef(map.things[hilight_item]->type)->args[2]);
		thing_args.add(get_tdef(map.things[hilight_item]->type)->args[3]);
		thing_args.add(get_tdef(map.things[hilight_item]->type)->args[4]);
	}
	else
	{
		for (int a = 0; a < 5; a++)
			thing_args.add("Arg" + a);
	}
}

// menu_draw: Draw the currently open menu (if any)
// --------------------------------------------- >>
void menu_draw()
{
	if (!menu)
		return;

	// Things
	if (open_menu == MENU_THINGMAIN)
		gui_popupmenu(thing_options, "Thing Properties");

	if (open_menu == MENU_THINGANGLE)
		gui_popupmenu(thing_angles, "Change Angle");

	if (open_menu == MENU_THINGFLAGS)
		gui_popupmenu_toggle(thing_flags, "Change Flags");

	if (open_menu == MENU_THINGTYPEGROUPS)
		gui_popupmenu(thing_groups, "Select Category");

	if (open_menu == MENU_THINGTYPESELECT)
		gui_popupmenu(thing_types[thing_group], "Select Type");

	if (open_menu == MENU_THINGARGS)
		gui_popupmenu(thing_args, "Change Args");


	// Lines
	if (open_menu == MENU_LINEMAIN)
		gui_popupmenu(line_options, "Line Properties");

	if (open_menu == MENU_LINEFLAGS)
		gui_popupmenu_toggle(line_flags, "Change Flags");

	if (open_menu == MENU_LINETYPEGROUPS)
		gui_popupmenu(line_groups, "Select Category");

	if (open_menu == MENU_LINETYPESELECT)
		gui_popupmenu(line_types[line_group], "Select Special");

	if (open_menu == MENU_LINEARGS)
		gui_popupmenu(line_args, "Change Args");

	if (open_menu == MENU_LINETRIGGERS)
		gui_popupmenu(line_triggers, "Change Trigger");

	if (open_menu == MENU_LINEARGTYPE)
		gui_popupmenu(line_arg_type, "Change Argument");


	// Sides
	if (open_menu == MENU_SIDEMAIN)
		gui_popupmenu(side_options, "Edit Side %d", line_side);


	// Sectors
	if (open_menu == MENU_SECTORMAIN)
		gui_popupmenu(sector_options, "Sector Properties");

	if (open_menu == MENU_SECTORTYPESELECT)
		gui_popupmenu(sector_typelist, "Select Special");


	// Main menus
	if (open_menu == MENU_FILEMENU)
		gui_drawmenu(file_menu, 8, 16);

	if (open_menu == MENU_EDITMENU)
		gui_drawmenu(edit_menu, 48, 16);

	if (open_menu == MENU_OPTIONSMENU)
		gui_drawmenu(options_menu, 88, 16);

	if (open_menu == MENU_MODEMENU)
		gui_drawmenu(mode_menu, 152, 16);

	if (open_menu == MENU_CHECKMENU)
		gui_drawmenu(check_menu, 192, 16);

	// Misc menus
	if (open_menu == MENU_COLOURSCHEMES)
		gui_popupmenu(colour_schemes, "Select Colour Scheme:");

	if (open_menu == MENU_SKINS)
		gui_popupmenu(skin_dirs, "Select UI Skin:");

	// Options menus
	if (open_menu == MENU_DISPLAY)
		gui_popupmenu(display_menu, "Display Options");

	if (open_menu == MENU_VIDEO)
		gui_popupmenu(video_menu, "Set Video Mode");

	if (open_menu == MENU_FILTER)
		gui_popupmenu(filter_menu, "Set GL Filter");

	if (open_menu == MENU_BROWSER)
		gui_popupmenu(browser_menu, "Browser Options");

	if (open_menu == MENU_BROWSER_ORDER)
		gui_popupmenu(browser_order_menu, "Set Browser Order");

	// Control menus
	if (open_menu == MENU_CONTROLS)
		gui_popupmenu(controls_menu, "Customize Controls");

	if (open_menu == MENU_2DCONTROLS)
		gui_popupmenu(controls_2dmode, "Customize 2dmode Controls");

	if (open_menu == MENU_3DCONTROLS)
		gui_popupmenu(controls_3dmode, "Customize 3dmode Controls");

	if (open_menu == MENU_MISCCONTROLS)
		gui_popupmenu(controls_misc, "Customize Misc Controls");

	if (inbox)
	{
		if (open_menu == MENU_THINGMAIN)
		{
			if (sel_item == 3)
				gui_inputbox("Change Thing ID", "Enter A Number (-1 For Lowest Free):");

			if (sel_item == 6)
				gui_inputbox("Change Thing Z Height", "Enter A Number:");
		}

		if (open_menu == MENU_LINEMAIN)
		{
			if (sel_item == 2)
				gui_inputbox("Change Sector Tag", "Enter A Number (-1 For Lowest Free):");
		}

		if (open_menu == MENU_LINEARGS)
			gui_inputbox("Change Line Arg", "Enter Argument Value");

		if (open_menu == MENU_SIDEMAIN)
		{
			if (sel_item == 0)
				gui_inputbox("Change X Offset", "Enter New X Offset:");
			if (sel_item == 1)
				gui_inputbox("Change Y Offset", "Enter New Y Offset:");
			if (sel_item == 5)
				gui_inputbox("Change Sector Reference", "Enter Sector:");
		}

		if (open_menu == MENU_SECTORMAIN)
		{
			if (sel_item == 0)
				gui_inputbox("Change Floor Height", "Enter New Floor Height:");
			if (sel_item == 1)
				gui_inputbox("Change Ceiling Height", "Enter New Ceiling Height:");
			if (sel_item == 4)
				gui_inputbox("Change Light Level", "Enter New Light Level (0-255):");
			if (sel_item == 6)
				gui_inputbox("Change Tag", "Enter A Number (-1 For Lowest Free):");
		}

		if (open_menu == MENU_THINGARGS)
			gui_inputbox("Change Thing Arg", "Enter Argument Value");

		if (open_menu == MENU_THINGTYPEGROUPS)
		{
			if (sel_item == thing_groups.n_strings - 1)
				gui_inputbox("Change Thing Type", "Enter Type Number:");
		}

		if (open_menu == MENU_LINETYPEGROUPS)
		{
			if (sel_item == line_groups.n_strings - 1)
				gui_inputbox("Change Line Special", "Enter Line Special:");
		}

		if (open_menu == MENU_FILEMENU)
		{
			if (sel_item == 1)
				gui_inputbox("Save Map As...", "Enter Map Filename:");
		}

		if (open_menu == MENU_EDITMENU)
		{
			if (sel_item == 1)
			{
				if (edit_mode == 1)
					gui_inputbox("Extrude Amount", "Enter Distance To Extrude");

				if (edit_mode == 2)
					gui_inputbox("Stairs Height", "Enter Step height (- for downwards)");
			}

			if (sel_item == 2)
			{
				if (edit_mode == 2)
					gui_inputbox("Stairs Height", "Enter Step height (- for downwards)");
			}
		}

		if (open_menu == MENU_DISPLAY)
		{
			if (sel_item == 3)
				gui_inputbox("Set Line Width", "Enter Width:");
		}

		if (open_menu == MENU_BROWSER)
		{
			if (sel_item == 0)
				gui_inputbox("Set Browser Columns", "Enter Number (Minimum 4):");
		}
	}
}

// get_line_flags: Sets up the line flags menu with the currently hilighted line's flags
// ---------------------------------------------------------------------------------- >>
void get_line_flags()
{
	WORD flags = map.lines[hilight_item]->flags;

	for (int s = 0; s < 16; s++)
		sel_items[s] = false;

	if (flags & LINE_IMPASSIBLE)
		sel_items[0] = true;

	if (flags & LINE_BLOCKMONSTERS)
		sel_items[1] = true;

	if (flags & LINE_TWOSIDED)
		sel_items[2] = true;

	if (flags & LINE_UPPERUNPEGGED)
		sel_items[3] = true;

	if (flags & LINE_LOWERUNPEGGED)
		sel_items[4] = true;

	if (flags & LINE_SECRET)
		sel_items[5] = true;

	if (flags & LINE_BLOCKSOUND)
		sel_items[6] = true;

	if (flags & LINE_NOTONMAP)
		sel_items[7] = true;

	if (flags & LINE_STARTONMAP)
		sel_items[8] = true;

	if (flags & LINE_REPEATABLE)
		sel_items[9] = true;

	if (flags & LINE_MONSTERCANACT)
		sel_items[10] = true;

	if (flags & LINE_BLOCKALL)
		sel_items[11] = true;
}

// get_thing_flags: Sets up the thing flags menu with the currently hilighted thing's flags
// ------------------------------------------------------------------------------------- >>
void get_thing_flags()
{
	if (hilight_item == -1)
		return;

	WORD flags = map.things[hilight_item]->flags;

	for (int s = 0; s < 16; s++)
		sel_items[s] = false;

	if (flags & THING_EASY)
		sel_items[0] = true;

	if (flags & THING_MEDIUM)
		sel_items[1] = true;

	if (flags & THING_HARD)
		sel_items[2] = true;

	if (flags & THING_DEAF)
		sel_items[3] = true;

	if (!map.hexen)
	{
		if (flags & THING_MULTI)
			sel_items[4] = true;
	}
	else
	{
		if (flags & THING_DORMANT)
			sel_items[4] = true;

		if (flags & THING_FIGHTER)
			sel_items[5] = true;

		if (flags & THING_CLERIC)
			sel_items[6] = true;

		if (flags & THING_MAGE)
			sel_items[7] = true;

		if (flags & THING_SINGLE)
			sel_items[8] = true;

		if (flags & THING_COOPERATIVE)
			sel_items[9] = true;

		if (flags & THING_DEATHMATCH)
			sel_items[10] = true;

		if (map.zdoom)
		{
			if (flags & THING_TRANSLUCENT)
				sel_items[11] = true;

			if (flags & THING_INVISIBLE)
				sel_items[12] = true;

			if (flags & THING_FRIENDLY)
				sel_items[13] = true;

			if (flags & THING_STANDSTILL)
				sel_items[14] = true;
		}
	}
}

// toggle_line_flag: Applies selected line flags to all selected/hilighted line(s)
// ---------------------------------------------------------------------------- >>
void toggle_line_flag()
{
	DWORD newflags = 0;

	if (sel_items[0])
		newflags |= LINE_IMPASSIBLE;
	if (sel_items[1])
		newflags |= LINE_BLOCKMONSTERS;
	if (sel_items[2])
		newflags |= LINE_TWOSIDED;
	if (sel_items[3])
		newflags |= LINE_UPPERUNPEGGED;
	if (sel_items[4])
		newflags |= LINE_LOWERUNPEGGED;
	if (sel_items[5])
		newflags |= LINE_SECRET;
	if (sel_items[6])
		newflags |= LINE_BLOCKSOUND;
	if (sel_items[7])
		newflags |= LINE_NOTONMAP;
	if (sel_items[8])
		newflags |= LINE_STARTONMAP;
	if (sel_items[9])
		newflags |= LINE_REPEATABLE;
	if (sel_items[10])
		newflags |= LINE_MONSTERCANACT;
	if (sel_items[11])
		newflags |= LINE_BLOCKALL;

	bool select = selection();

	/*
	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (map.lines[l]->selected)
			map.lines[l]->flags = newflags;

		if (!select && map.lines[l]->hilighted)
			map.lines[l]->flags = newflags;
	}
	*/
	if (selection())
	{
		for (DWORD a = 0; a < selected_items.n_numbers; a++)
			map.lines[selected_items.numbers[a]]->flags = newflags;
	}
	else if (hilight_item != -1)
		map.lines[hilight_item]->flags = newflags;
}

// toggle_thing_flag: Applies selected thing flags to all selected/hilighted thing(s)
// ------------------------------------------------------------------------------- >>
void toggle_thing_flag()
{
	short newflags = 0;

	if (sel_items[0])
		newflags |= THING_EASY;
	if (sel_items[1])
		newflags |= THING_MEDIUM;
	if (sel_items[2])
		newflags |= THING_HARD;
	if (sel_items[3])
		newflags |= THING_DEAF;

	if (!map.hexen)
	{
		if (sel_items[4])
			newflags |= THING_MULTI;
	}
	else
	{
		if (sel_items[4])
			newflags |= THING_DORMANT;
		if (sel_items[5])
			newflags |= THING_FIGHTER;
		if (sel_items[6])
			newflags |= THING_CLERIC;
		if (sel_items[7])
			newflags |= THING_MAGE;
		if (sel_items[8])
			newflags |= THING_SINGLE;
		if (sel_items[9])
			newflags |= THING_COOPERATIVE;
		if (sel_items[10])
			newflags |= THING_DEATHMATCH;

		if (map.zdoom)
		{
			if (sel_items[11])
				newflags |= THING_TRANSLUCENT;
			if (sel_items[12])
				newflags |= THING_INVISIBLE;
			if (sel_items[13])
				newflags |= THING_FRIENDLY;
			if (sel_items[14])
				newflags |= THING_STANDSTILL;
		}
	}

	bool select = selection();

	/*
	for (DWORD t = 0; t < map.n_things; t++)
	{
		if (map.things[t]->selected)
			map.things[t]->flags = newflags;

		if (!select && map.things[t]->hilighted)
			map.things[t]->flags = newflags;
	}
	*/
	if (selection())
	{
		for (DWORD a = 0; a < selected_items.n_numbers; a++)
			map.things[selected_items.numbers[a]]->flags = newflags;
	}
	else if (hilight_item != -1)
		map.things[hilight_item]->flags = newflags;

	last_thing.flags = newflags;
}

// set_thing_angle: Sets the angle of all hilighted/selected things
// ------------------------------------------------------------- >>
void set_thing_angle()
{
	WORD newangle = 0;

	switch (sel_item)
	{
	case 0:
		newangle = 0;
		break;
	case 1:
		newangle = 45;
		break;
	case 2:
		newangle = 90;
		break;
	case 3:
		newangle = 135;
		break;
	case 4:
		newangle = 180;
		break;
	case 5:
		newangle = 225;
		break;
	case 6:
		newangle = 270;
		break;
	case 7:
		newangle = 315;
		break;
	}

	bool select = selection();

	if (selection())
	{
		for (DWORD a = 0; a < selected_items.n_numbers; a++)
			map.things[selected_items.numbers[a]]->angle = newangle;
	}
	else if (hilight_item != -1)
		map.things[hilight_item]->angle = newangle;

	last_thing.angle = newangle;
}

// menu_action: User has selected a menu item
// --------------------------------------- >>
bool menu_action()
{
	// << ---------------------- >>
	// << ----- Thing Menu ----- >>
	// << ---------------------- >>

	// Main Thing Menu
	if (open_menu == MENU_THINGMAIN)
	{
		switch (sel_item)
		{
		case 0: // Change Type
			open_menu = MENU_THINGTYPEGROUPS;
			sel_item = 0;
			max_item = thing_groups.n_strings - 1;
			break;
		case 1: // Change Angle
			open_menu = MENU_THINGANGLE;
			sel_item = 0;
			max_item = thing_angles.n_strings - 1;
			break;
		case 2: // Change Flags
			open_menu = MENU_THINGFLAGS;
			sel_item = 0;
			max_item = thing_flags.n_strings - 1;
			get_thing_flags();
			break;
		case 3: // Change TID or Browse Things
			if (map.hexen)
			{
				if (hilight_item != -1)
				{
					char temp[8] = "";
					sprintf(temp, "%d", map.things[hilight_item]->tid);
					menu_input = temp;
				}
				inbox_typed = false;
				inbox = true;
				break;
			}
			else
			{
				if (!browser)
				{
					setup_browser_things();

					if (!selection() && hilight_item != -1)
						browser_findtex(get_tdef(map.things[hilight_item]->type)->name);

					browser = true;
					break;
				}
				else
				{
					if (selection())
					{
						for (DWORD a = 0; a < selected_items.n_numbers; a++)
							map.things[selected_items.numbers[a]]->type = browser_getthing();
					}
					else if (hilight_item != -1)
						map.things[hilight_item]->type = browser_getthing();

					last_thing.type = browser_getthing();
					update_map_things();
					update_map();

					break;
				}
			}
		case 4: // Change Special
			open_menu = MENU_LINETYPEGROUPS;
			setup_thing_args();
			sel_item = 0;
			max_item = line_groups.n_strings - 1;
			break;
		case 5: // Change Args
			open_menu = MENU_THINGARGS;
			setup_thing_args();
			sel_item = 0;
			max_item = thing_args.n_strings - 1;
			break;
		case 6: // Change Z Height
			inbox = true;
			inbox_typed = false;
			break;
		case 7: // Browse Things
			if (!browser)
			{
				setup_browser_things();

				if (!selection() && hilight_item != -1)
					browser_findtex(get_tdef(map.things[hilight_item]->type)->name);

				browser = true;
				break;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.things[selected_items.numbers[a]]->type = browser_getthing();
				}
				else if (hilight_item != -1)
					map.things[hilight_item]->type = browser_getthing();

				last_thing.type = browser_getthing();
				update_map_things();
				update_map();

				break;
			}
		}

		return true;
	}

	// Thing type groups menu
	if (open_menu == MENU_THINGTYPEGROUPS)
	{
		if (sel_item == thing_groups.n_strings - 1)
		{
			// Enter manual thing type
			inbox = true;
			inbox_typed = false;
		}
		else
		{
			thing_group = sel_item;
			sel_item = 0;
			max_item = thing_types[thing_group].n_strings - 1;
			open_menu = MENU_THINGTYPESELECT;
			return true;
		}
	}

	// Thing type select menu
	if (open_menu == MENU_THINGTYPESELECT)
	{
		bool select = selection();
		DWORD thing = 0;

		if (selection())
		{
			for (DWORD a = 0; a < selected_items.n_numbers; a++)
			{
				thing = selected_items.numbers[a];
				thing_settype(thing_group, sel_item, thing);
			}
		}
		else if (hilight_item != -1)
		{
			thing = hilight_item;
			thing_settype(thing_group, sel_item, thing);
		}

		last_thing.type = map.things[thing]->type;

		map_changelevel(2);
		update_map_things();

		sel_item = 0;
		max_item = thing_options.n_strings - 1;
		open_menu = MENU_THINGMAIN;
		return true;
	}

	// Thing flags menu
	if (open_menu == MENU_THINGFLAGS)
	{
		sel_items[sel_item] = !sel_items[sel_item];
		toggle_thing_flag();
		map_changelevel(1);
		return true;
	}

	// Thing angle menu
	if (open_menu == MENU_THINGANGLE)
	{
		set_thing_angle();
		open_menu = MENU_THINGMAIN;
		max_item = thing_options.n_strings - 1;
		sel_item = 0;
		map_changelevel(1);
		return true;
	}

	// Thing args menu
	if (open_menu == MENU_THINGARGS)
	{
		if (hilight_item != -1)
		{
			char temp[8] = "";
			sprintf(temp, "%d", map.things[hilight_item]->args[sel_item]);
			menu_input = temp;
		}
		inbox_typed = false;
		inbox = true;
	}

	// << --------------------- >>
	// << ----- Line Menu ----- >>
	// << --------------------- >>

	// Main Line Menu
	if (open_menu == MENU_LINEMAIN)
	{
		switch (sel_item)
		{
		case 0:
			// Edit First side
			if (hilight_item != -1 && map.lines[hilight_item]->side1 != -1)
			{
				open_menu = MENU_SIDEMAIN;
				line_side = 1;
				sel_item = 0;
				max_item = side_options.n_strings - 1;
			}
			// Add first side
			if (hilight_item != -1 && map.lines[hilight_item]->side1 == -1)
			{
				int side = 0;

				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
					{
						side = map.add_side();
						map.lines[selected_items.numbers[a]]->side1 = side;
					}
				}
				else if (hilight_item != -1)
				{
					side = map.add_side();
					map.lines[hilight_item]->side1 = side;
				}

				map_changelevel(2);
				setup_line_menu();
			}
			break;
		case 1:
			// Edit Second side
			if (hilight_item != -1 && map.lines[hilight_item]->side2 != -1)
			{
				open_menu = MENU_SIDEMAIN;
				line_side = 2;
				sel_item = 0;
				max_item = side_options.n_strings - 1;
			}
			// Add Second side
			if (hilight_item != -1 && map.lines[hilight_item]->side2 == -1)
			{
				int side = 0;

				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
					{
						side = map.add_side();
						map.lines[selected_items.numbers[a]]->side2 = side;
						map.lines[selected_items.numbers[a]]->set_flag(LINE_TWOSIDED);
						map.lines[selected_items.numbers[a]]->clear_flag(LINE_IMPASSIBLE);
					}
				}
				else if (hilight_item != -1)
				{
					side = map.add_side();
					map.lines[hilight_item]->side2 = side;
					map.lines[hilight_item]->set_flag(LINE_TWOSIDED);
					map.lines[hilight_item]->clear_flag(LINE_IMPASSIBLE);
				}

				map_changelevel(2);
				setup_line_menu();
			}
			break;
		case 2: // Change Sector Tag
			if (hilight_item != -1)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.lines[hilight_item]->sector_tag);
				menu_input = temp;
			}
			inbox_typed = false;
			inbox = true;
			break;
		case 3: // Change Type
			open_menu = MENU_LINETYPEGROUPS;
			sel_item = 0;
			max_item = line_groups.n_strings - 1;
			break;
		case 4: // Change Flags
			open_menu = MENU_LINEFLAGS;
			sel_item = 0;
			max_item = line_flags.n_strings - 1;
			get_line_flags();
			break;
		case 5: // Change Args
			open_menu = MENU_LINEARGS;
			setup_line_args();
			sel_item = 0;
			max_item = line_args.n_strings - 1;
			break;
		case 6: // Change Trigger
			open_menu = MENU_LINETRIGGERS;
			if (hilight_item != -1)
				sel_item = map.lines[hilight_item]->get_trigger();
			else
				sel_item = 0;
			max_item = line_triggers.n_strings - 1;
			break;
		}

		return true;
	}

	// Line flags menu
	if (open_menu == MENU_LINEFLAGS)
	{
		sel_items[sel_item] = !sel_items[sel_item];
		toggle_line_flag();
		map_changelevel(2);
		return true;
	}

	// Line args menu
	if (open_menu == MENU_LINEARGS)
	{
		if (hilight_item != -1)
		{
			if (get_ltype(map.lines[hilight_item]->type)->arg_types[sel_item] == NULL)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.lines[hilight_item]->args[sel_item]);
				menu_input = temp;
				inbox_typed = false;
				inbox = true;
			}
			else
			{
				argtype_t *type = get_argtype(*(get_ltype(map.lines[hilight_item]->type)->arg_types[sel_item]));
				line_arg_type.clear();

				for (DWORD t = 0; t < type->n_values; t++)
					line_arg_type.add(type->values[t]->name);

				open_menu = MENU_LINEARGTYPE;
				sel_arg = sel_item;
				sel_item = 0;
				max_item = line_arg_type.n_strings - 1;
				return true;
			}
		}
		inbox_typed = false;
		inbox = true;
		return true;
	}

	// Line trigger menu
	if (open_menu == MENU_LINETRIGGERS)
	{
		if (selection())
		{
			for (DWORD a = 0; a < selected_items.n_numbers; a++)
				map.lines[selected_items.numbers[a]]->set_trigger(sel_item);
		}
		else if (hilight_item != -1)
			map.lines[hilight_item]->set_trigger(sel_item);

		sel_item = 0;
		max_item = line_options.n_strings - 1;
		open_menu = MENU_LINEMAIN;
		return true;
	}

	// Line type groups menu
	if (open_menu == MENU_LINETYPEGROUPS)
	{
		if (sel_item > 0 && sel_item != line_groups.n_strings - 1)
		{
			line_group = sel_item - 1;
			sel_item = 0;
			max_item = line_types[line_group].n_strings - 1;
			open_menu = MENU_LINETYPESELECT;
		}
		else if (sel_item == 0) // No special selected
		{
			bool select = selection();

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.lines[selected_items.numbers[a]]->type = 0;
			}
			else if (hilight_item != -1)
				map.lines[hilight_item]->type = 0;

			sel_item = 0;
			max_item = line_options.n_strings - 1;
			open_menu = MENU_LINEMAIN;
		}
		else
		{
			// Enter special manually
			inbox = true;
			inbox_typed = false;
		}

		return true;
	}

	// Line type select menu
	if (open_menu == MENU_LINETYPESELECT)
	{
		bool select = selection();

		if (edit_mode == 1)
		{
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					line_settype(line_group, sel_item, selected_items.numbers[a]);
			}
			else if (hilight_item != -1)
				line_settype(line_group, sel_item, hilight_item);

			sel_item = 0;
			max_item = line_options.n_strings - 1;
			open_menu = MENU_LINEMAIN;
			return true;
		}

		if (map.zdoom && edit_mode == 3)
		{
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					thing_setspecial(line_group, sel_item, selected_items.numbers[a]);
			}
			else if (hilight_item != -1)
				thing_setspecial(line_group, sel_item, hilight_item);

			sel_item = 0;
			max_item = thing_options.n_strings - 1;
			open_menu = MENU_THINGMAIN;
			return true;
		}
	}

	// Line arg type menu
	if (open_menu == MENU_LINEARGTYPE)
	{
		if (hilight_item != -1)
		{
			argtype_t *type = get_argtype(*(get_ltype(map.lines[hilight_item]->type)->arg_types[sel_arg]));

			BYTE value = type->values[sel_item]->value;
			map.lines[hilight_item]->args[sel_arg] = value;

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.lines[selected_items.numbers[a]]->args[sel_arg] = value;
			}

			open_menu = MENU_LINEARGS;
			sel_item = 0;
			max_item = line_args.n_strings - 1;
			return true;
		}
	}

	// << --------------------- >>
	// << ----- Side Menu ----- >>
	// << --------------------- >>

	if (open_menu == MENU_SIDEMAIN)
	{
		switch (sel_item)
		{
		case 0: // Change x offset
			if (hilight_item != -1)
			{
				char temp[8] = "";

				if (line_side == 1)
					sprintf(temp, "%d", map.l_getxoff(hilight_item, 1));
				else
					sprintf(temp, "%d", map.l_getxoff(hilight_item, 2));

				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		case 1: // Change y offset
			if (hilight_item != -1)
			{
				char temp[8] = "";

				if (line_side == 1)
					sprintf(temp, "%d", map.l_getyoff(hilight_item, 1));
				else
					sprintf(temp, "%d", map.l_getyoff(hilight_item, 2));

				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		case 2: // Change midtex
			if (!browser)
			{
				if (hilight_item != -1)
				{
					if (line_side == 1)
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 1)->tex_middle);
					else
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 2)->tex_middle);
				}
				else
					setup_browser(true, mix_tex, "-");

				browser = true;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.l_setmidtex(selected_items.numbers[a], line_side, browser_gettex());
				}
				else if (hilight_item != -1)
					map.l_setmidtex(hilight_item, line_side, browser_gettex());
			}
			break;
		case 3: // Change upper tex
			if (!browser)
			{
				if (hilight_item != -1)
				{
					if (line_side == 1)
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 1)->tex_upper);
					else
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 2)->tex_upper);
				}
				else
					setup_browser(true, mix_tex, "-");

				browser = true;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.l_setuptex(selected_items.numbers[a], line_side, browser_gettex());
				}
				else if (hilight_item != -1)
					map.l_setuptex(hilight_item, line_side, browser_gettex());
			}
			break;
		case 4: // Change lower tex
			if (!browser)
			{
				if (hilight_item != -1)
				{
					if (line_side == 1)
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 1)->tex_lower);
					else
						setup_browser(true, mix_tex, map.l_getside(hilight_item, 2)->tex_lower);
				}
				else
					setup_browser(true, mix_tex, "-");

				browser = true;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.l_setlotex(selected_items.numbers[a], line_side, browser_gettex());
				}
				else if (hilight_item != -1)
					map.l_setlotex(hilight_item, line_side, browser_gettex());
			}
			break;
		case 5: // Change sector ref
			if (hilight_item != -1)
			{
				char temp[8] = "";

				if (line_side == 1)
					sprintf(temp, "%d", map.l_getsector1(hilight_item));
				else
					sprintf(temp, "%d", map.l_getsector2(hilight_item));

				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		}

		return true;
	}

	// << ----------------------- >>
	// << ----- Sector Menu ----- >>
	// << ----------------------- >>

	if (open_menu == MENU_SECTORMAIN)
	{
		switch (sel_item)
		{
		case 0: // Floor height
			if (hilight_item != -1)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.sectors[hilight_item]->f_height);
				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		case 1: // Ceil height
			if (hilight_item != -1)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.sectors[hilight_item]->c_height);
				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		case 2: // Floor tex
			if (!browser)
			{
				if (hilight_item != -1)
					setup_browser(mix_tex, true, map.sectors[hilight_item]->f_tex);
				else
					setup_browser(mix_tex, true, "-");

				browser = true;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.sectors[selected_items.numbers[a]]->f_tex = browser_gettex();
				}
				else if (hilight_item != -1)
					map.sectors[hilight_item]->f_tex = browser_gettex();
			}
			break;
		case 3: // Ceil tex
			if (!browser)
			{
				if (hilight_item != -1)
					setup_browser(mix_tex, true, map.sectors[hilight_item]->c_tex);
				else
					setup_browser(mix_tex, true, "-");

				browser = true;
			}
			else
			{
				if (selection())
				{
					for (DWORD a = 0; a < selected_items.n_numbers; a++)
						map.sectors[selected_items.numbers[a]]->c_tex = browser_gettex();
				}
				else if (hilight_item != -1)
					map.sectors[hilight_item]->c_tex = browser_gettex();
			}
			break;
			break;
		case 4: // Light
			if (hilight_item != -1)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.sectors[hilight_item]->light);
				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		case 5: // Special
			open_menu = MENU_SECTORTYPESELECT;
			sel_item = 0;
			max_item = sector_typelist.n_strings - 1;
			break;
		case 6: // Tag
			if (hilight_item != -1)
			{
				char temp[8] = "";
				sprintf(temp, "%d", map.sectors[hilight_item]->tag);
				menu_input = temp;
			}
			inbox = true;
			inbox_typed = false;
			break;
		}

		return true;
	}

	// Sector type select menu
	if (open_menu == MENU_SECTORTYPESELECT)
	{
		bool select = selection();

		/*
		for (DWORD s = 0; s < map.n_sectors; s++)
		{
			if (open_map.SIsSelected(s))
			sector_settype(sel_item, s);

			if (!select && open_map.SIsHilighted(s))
			sector_settype(sel_item, s);
		}
		*/

		if (selection())
		{
			for (DWORD a = 0; a < selected_items.n_numbers; a++)
				sector_settype(sel_item, selected_items.numbers[a]);
		}
		else if (hilight_item != -1)
			sector_settype(sel_item, hilight_item);
			//map.sectors[hilight_item]->c_tex = browser_gettex();

		sel_item = 0;
		max_item = sector_options.n_strings - 1;
		open_menu = MENU_SECTORMAIN;
		return true;
	}


	// << ----------------------- >>
	// << ----- Menu Bar -------- >>
	// << ----------------------- >>

	// File menu
	if (open_menu == MENU_FILEMENU)
	{
		if (sel_item == 0)
		{
			if (wads.get_last())
			{
				map.add_to_wad(wads.get_last());
				wads.get_last()->save(true);
				menu = false;
			}
			else
				sel_item = 1;
		}

		if (sel_item == 1)
		{
			inbox = true;
			inbox_typed = false;
		}

		if (sel_item == 2)
			return false;
	}

	// Edit menu
	if (open_menu == MENU_EDITMENU)
	{
		if (sel_item == 0)
		{
			edit_item();
			menu = false;
		}

		if (sel_item == 1)
		{
			// Extrude lines
			if (edit_mode == 1)
			{
				inbox = true;
				inbox_typed = false;
				return true;
			}

			// Create Stairs
			if (edit_mode == 2)
			{
				inbox = true;
				inbox_typed = false;
				return true;
			}
		}

		if (sel_item == 2)
		{
			// Create Stairs Floor+Ceil
			if (edit_mode == 2)
			{
				inbox = true;
				inbox_typed = false;
				return true;
			}
		}

		if (sel_item == 3)
		{
			// Create Door
			if (edit_mode == 2)
			{
				if (!browser)
				{
					browser = true;
					setup_browser(true, mix_tex, "-");
				}
				else
					sector_create_door(browser_gettex());
			}
		}
	}

	// Options menu
	if (open_menu == MENU_OPTIONSMENU)
	{
		if (sel_item == 0)
		{
			open_menu = MENU_DISPLAY;
			sel_item = 0;
			setup_display_menu();
			max_item = display_menu.n_strings - 1;
			return true;
		}

		if (sel_item == 2)
		{
			open_menu = MENU_BROWSER;
			sel_item = 0;
			max_item = browser_menu.n_strings - 1;
			return true;
		}

		if (sel_item == 3)
		{
			open_menu = MENU_VIDEO;
			sel_item = 0;
			max_item = video_menu.n_strings - 1;
			return true;
		}

		if (sel_item == 4)
		{
			open_menu = MENU_CONTROLS;
			sel_item = 0;
			max_item = controls_menu.n_strings - 1;
			return true;
		}
	}

	// Browser menu
	if (open_menu == MENU_BROWSER)
	{
		if (sel_item == 0)
		{
			char tmp[5] = "";
			sscanf(tmp, "%d", browser_columns);
			menu_input = tmp;
			inbox_typed = false;
			inbox = true;
			return true;
		}

		if (sel_item == 1)
		{
			open_menu = MENU_BROWSER_ORDER;
			sel_item = 0;
			max_item = browser_order_menu.n_strings - 1;
			return true;
		}
	}

	// Browser order menu
	if (open_menu == MENU_BROWSER_ORDER)
	{
		browser_order = sel_item;
		open_menu = MENU_BROWSER;
		max_item = browser_menu.n_strings - 1;
		sel_item = 1;
	}

	// Display menu
	if (open_menu == MENU_DISPLAY)
	{
		if (sel_item == 0)
		{
			open_menu = MENU_COLOURSCHEMES;
			sel_item = 0;
			max_item = colour_schemes.n_strings - 1;
			return true;
		}

		if (sel_item == 1)
		{
			open_menu = MENU_SKINS;
			sel_item = 0;
			max_item = skin_dirs.n_strings - 1;
			return true;
		}

		if (sel_item == 2)
		{
			line_aa = !line_aa;
			setup_display_menu();
			update_map();
			return true;
		}

		if (sel_item == 3)
		{
			char tmp[5] = "";
			sscanf(tmp, "%1.2f", line_size);
			menu_input = tmp;
			inbox_typed = false;
			inbox = true;
			return true;
		}

		if (sel_item == 4)
		{
			grid_dashed = !grid_dashed;
			setup_display_menu();
			update_map();
			return true;
		}

		if (sel_item == 5)
		{
			thing_sprites = !thing_sprites;
			setup_display_menu();
			update_map();
			return true;
		}

		if (sel_item == 6)
		{
			open_menu = MENU_FILTER;
			sel_item = 0;
			max_item = filter_menu.n_strings - 1;
			return true;
		}

		if (sel_item == 7)
		{
			vid_fullscreen = !vid_fullscreen;
			init();
			return true;
		}
	}

	// Video menu
	if (open_menu == MENU_VIDEO)
	{
		int bpp, width, height;

		if (sel_item == 0 || sel_item == 2 || sel_item == 4 ||
			sel_item == 6 || sel_item == 8 || sel_item == 10)
			bpp = 16;

		if (sel_item == 1 || sel_item == 3 || sel_item == 5 ||
			sel_item == 7 || sel_item == 9 || sel_item == 11)
			bpp = 32;

		if (sel_item == 0 || sel_item == 1)
		{
			width = 800;
			height = 600;
		}

		if (sel_item == 2 || sel_item == 3)
		{
			width = 1024;
			height = 768;
		}

		if (sel_item == 4 || sel_item == 5)
		{
			width = 1152;
			height = 864;
		}

		if (sel_item == 6 || sel_item == 7)
		{
			width = 1280;
			height = 960;
		}

		if (sel_item == 8 || sel_item == 9)
		{
			width = 1280;
			height = 1024;
		}

		if (sel_item == 10 || sel_item == 11)
		{
			width = 1600;
			height = 1200;
		}

		if (vid_bitdepth != bpp || vid_width != width || vid_height != height)
		{
			vid_width = width;
			vid_height = height;
			vid_bitdepth = bpp;
			init();
		}

		menu = false;
	}

	// Controls menu
	if (open_menu == MENU_CONTROLS)
	{
		if (sel_item == 0)
		{
			open_menu = MENU_2DCONTROLS;
			sel_item = 0;
			max_item = controls_2dmode.n_strings - 1;
			return true;
		}

		if (sel_item == 1)
		{
			open_menu = MENU_3DCONTROLS;
			sel_item = 0;
			max_item = controls_3dmode.n_strings - 1;
			return true;
		}

		if (sel_item == 2)
		{
			open_menu = MENU_MISCCONTROLS;
			sel_item = 0;
			max_item = controls_misc.n_strings - 1;
			return true;
		}
	}

	// Filter menu
	if (open_menu == MENU_FILTER)
	{
		int filter = 0;

		if (sel_item == 0)
			filter = 1;

		if (sel_item == 1)
			filter = 2;

		if (sel_item == 2)
			filter = 3;

		if (gl_filter != filter)
		{
			gl_filter = filter;
			init();
		}

		menu = false;
	}

	// Checks menu
	if (open_menu == MENU_CHECKMENU)
	{
		if (sel_item == 0)
		{
			char temp[64] = "";

			sprintf(temp, "Stats for %s", map.name.c_str());
			message_title = temp;

			sprintf(temp, "Vertices: %d", map.n_verts);
			popup_message.add(temp);
			sprintf(temp, "Lines: %d", map.n_lines);
			popup_message.add(temp);
			sprintf(temp, "Sides: %d", map.n_sides);
			popup_message.add(temp);
			sprintf(temp, "Sectors: %d", map.n_sectors);
			popup_message.add(temp);
			sprintf(temp, "Things: %d", map.n_things);
			popup_message.add(temp);

			message = true;
			menu = false;
		}

		if (sel_item == 1)
		{
			char temp[64] = "";

			sprintf(temp, "Found %d unmatched tags.", clean_tags());
			popup_message.add(temp);
			message_title = "Clean Tags";
			message = true;
			menu = false;
			map_changelevel(1);
		}

		if (sel_item == 2)
		{
			char temp[64] = "";

			sprintf(temp, "Removed %d unattached vertices.", remove_free_verts());
			popup_message.add(temp);
			message_title = "Remove Free Vertices";
			message = true;
			menu = false;
			map_changelevel(2);
		}

		if (sel_item == 3)
		{
			char temp[64] = "";

			sprintf(temp, "Found %d zero-length lines.", remove_zerolength_lines());
			remove_free_verts();
			popup_message.add(temp);
			message_title = "Remove 0-Length Lines";
			message = true;
			menu = false;
			map_changelevel(2);
		}

		if (sel_item == 4)
		{
			char temp[64] = "";

			sprintf(temp, "Found %d unused sectors.", remove_unused_sectors());
			popup_message.add(temp);
			message_title = "Remove Unused Sectors";
			message = true;
			menu = false;
			map_changelevel(2);
		}
	}

	// Mode menu
	if (open_menu == MENU_MODEMENU)
	{
		if (sel_item < 4)
		{
			edit_mode = sel_item;
			update_map();
			menu = false;
		}
	}

	// << ------------------------- >>
	// << ----- Misc Menus -------- >>
	// << ------------------------- >>

	if (open_menu == MENU_COLOURSCHEMES)
		set_colour_config(*colour_schemes.strings[sel_item]);

	if (open_menu == MENU_SKINS)
	{
		ui_skin = *skin_dirs.strings[sel_item];
		//init();
		load_editor_textures();
	}

	return true;
}

// menu_inbox_action: User has entered a value to the input box
// --------------------------------------------------------- >>
void menu_inbox_action()
{
	if (open_menu == MENU_THINGMAIN)
	{
		if (sel_item == 3)
		{
			bool select = selection();
			int tid = atoi(menu_input.c_str());

			if (tid == -1)
				tid = get_free_tid();

			/*
			for (DWORD t = 0; t < map.n_things; t++)
			{
				if (SELECTED(things, t))
					map.things[t]->tid = tid;

				if (!select && HILIGHTED(things, t))
					map.things[t]->tid = tid;
			}
			*/

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.things[selected_items.numbers[a]]->tid = tid;
			}
			else if (hilight_item != -1)
				map.things[hilight_item]->tid = tid;
		}

		if (sel_item == 6)
		{
			int z = atoi(menu_input.c_str());

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.things[selected_items.numbers[a]]->z = z;
			}
			else if (hilight_item != -1)
				map.things[hilight_item]->z = z;
		}
	}

	if (open_menu == MENU_LINEMAIN)
	{
		if (sel_item == 2)
		{
			bool select = selection();
			int tag = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &tag);

			if (tag == -1)
				tag = get_free_tag();

			/*
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				if (SELECTED(lines, l))
					map.lines[l]->sector_tag = tag;

				if (!select && HILIGHTED(lines, l))
					map.lines[l]->sector_tag = tag;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.lines[selected_items.numbers[a]]->sector_tag = tag;
			}
			else if (hilight_item != -1)
				map.lines[hilight_item]->sector_tag = tag;
		}
	}

	if (open_menu == MENU_LINEARGS)
	{
		bool select = selection();
		int arg = atoi(menu_input.c_str());

		/*
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (SELECTED(lines, l))
				map.lines[l]->args[sel_item] = arg;

			if (!select && HILIGHTED(lines, l))
				map.lines[l]->args[sel_item] = arg;
		}
		*/
		if (selection())
		{
			for (DWORD a = 0; a < selected_items.n_numbers; a++)
				map.lines[selected_items.numbers[a]]->args[sel_item] = arg;
		}
		else if (hilight_item != -1)
			map.lines[hilight_item]->args[sel_item] = arg;
	}

	if (open_menu == MENU_SIDEMAIN)
	{
		if (sel_item == 0) // Change x offset
		{
			bool select = selection();
			int xoff = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &xoff);

			/*
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				if (SELECTED(lines, l))
				{
					if (line_side == 1)
					{
						if (map.lines[l]->side1 != -1)
							map.sides[map.lines[l]->side1]->x_offset = xoff;
					}
					else
					{
						if (map.lines[l]->side2 != -1)
							map.sides[map.lines[l]->side2]->x_offset = xoff;
					}
				}

				if (!select && HILIGHTED(lines, l))
				{
					if (line_side == 1)
						map.sides[map.lines[l]->side1]->x_offset = xoff;
					else
						map.sides[map.lines[l]->side2]->x_offset = xoff;
				}
			}
			*/

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
				{
					DWORD l = selected_items.numbers[a];

					if (line_side == 1 && map.lines[l]->side1 != -1)
						map.sides[map.lines[l]->side1]->x_offset = xoff;
					else if (line_side == 2 && map.lines[l]->side2 != -1)
						map.sides[map.lines[l]->side2]->x_offset = xoff;
				}
			}
			else if (hilight_item != -1)
			{
				if (line_side == 1 && map.lines[hilight_item]->side1 != -1)
					map.sides[map.lines[hilight_item]->side1]->x_offset = xoff;
				else if (line_side == 2 && map.lines[hilight_item]->side2 != -1)
					map.sides[map.lines[hilight_item]->side2]->x_offset = xoff;
			}

			map_changelevel(2);
		}

		if (sel_item == 1) // Change y offset
		{
			bool select = selection();
			int yoff = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &yoff);

			/*
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				if (SELECTED(lines, l))
				{
					if (line_side == 1)
					{
						if (map.lines[l]->side1 != -1)
							map.sides[map.lines[l]->side1]->y_offset = yoff;
					}
					else
					{
						if (map.lines[l]->side2 != -1)
							map.sides[map.lines[l]->side2]->y_offset = yoff;
					}
				}

				if (!select && HILIGHTED(lines, l))
				{
					if (line_side == 1)
						map.sides[map.lines[l]->side1]->y_offset = yoff;
					else
						map.sides[map.lines[l]->side2]->y_offset = yoff;
				}
			}
			*/

			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
				{
					DWORD l = selected_items.numbers[a];

					if (line_side == 1 && map.lines[l]->side1 != -1)
						map.sides[map.lines[l]->side1]->y_offset = yoff;
					else if (line_side == 2 && map.lines[l]->side2 != -1)
						map.sides[map.lines[l]->side2]->y_offset = yoff;
				}
			}
			else if (hilight_item != -1)
			{
				if (line_side == 1 && map.lines[hilight_item]->side1 != -1)
					map.sides[map.lines[hilight_item]->side1]->y_offset = yoff;
				else if (line_side == 2 && map.lines[hilight_item]->side2 != -1)
					map.sides[map.lines[hilight_item]->side2]->y_offset = yoff;
			}

			map_changelevel(2);
		}

		if (sel_item == 5) // Change sector ref
		{
			bool select = selection();
			int sector = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &sector);

			/*
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				if (SELECTED(lines, l))
				{
					if (line_side == 1)
					{
						if (map.lines[l]->side1 != -1)
							map.sides[map.lines[l]->side1]->sector = sector;
					}
					else
					{
						if (map.lines[l]->side2 != -1)
							map.sides[map.lines[l]->side2]->sector = sector;
					}
				}

				if (!select && HILIGHTED(lines, l))
				{
					if (line_side == 1)
						map.sides[map.lines[l]->side1]->sector = sector;
					else
						map.sides[map.lines[l]->side2]->sector = sector;
				}
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
				{
					DWORD l = selected_items.numbers[a];

					if (line_side == 1 && map.lines[l]->side1 != -1)
						map.sides[map.lines[l]->side1]->sector = sector;
					else if (line_side == 2 && map.lines[l]->side2 != -1)
						map.sides[map.lines[l]->side2]->sector = sector;
				}
			}
			else if (hilight_item != -1)
			{
				if (line_side == 1 && map.lines[hilight_item]->side1 != -1)
					map.sides[map.lines[hilight_item]->side1]->sector = sector;
				else if (line_side == 2 && map.lines[hilight_item]->side2 != -1)
					map.sides[map.lines[hilight_item]->side2]->sector = sector;
			}

			map_changelevel(2);
		}
	}

	if (open_menu == MENU_SECTORMAIN)
	{
		if (sel_item == 0)
		{
			bool select = selection();
			int height = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &height);

			/*
			for (DWORD s = 0; s < map.n_sectors; s++)
			{
				if (SELECTED(sectors, s))
					map.sectors[s]->f_height = height;

				if (!select && HILIGHTED(sectors, s))
					map.sectors[s]->f_height = height;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.sectors[selected_items.numbers[a]]->f_height = height;
			}
			else if (hilight_item != -1)
				map.sectors[hilight_item]->f_height = height;

			map_changelevel(2);
		}

		if (sel_item == 1)
		{
			bool select = selection();
			int height = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &height);

			/*
			for (DWORD s = 0; s < map.n_sectors; s++)
			{
				if (SELECTED(sectors, s))
					map.sectors[s]->c_height = height;

				if (!select && HILIGHTED(sectors, s))
					map.sectors[s]->c_height = height;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.sectors[selected_items.numbers[a]]->c_height = height;
			}
			else if (hilight_item != -1)
				map.sectors[hilight_item]->c_height = height;

			map_changelevel(2);
		}

		if (sel_item == 4)
		{
			bool select = selection();
			int light = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &light);

			if (light > 255)
				light = 255;

			if (light < 0)
				light = 0;

			/*
			for (DWORD s = 0; s < map.n_sectors; s++)
			{
				if (SELECTED(sectors, s))
					map.sectors[s]->light = light;

				if (!select && HILIGHTED(sectors, s))
					map.sectors[s]->light = light;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.sectors[selected_items.numbers[a]]->light = light;
			}
			else if (hilight_item != -1)
				map.sectors[hilight_item]->light = light;

			map_changelevel(2);
		}

		if (sel_item == 6)
		{
			bool select = selection();
			int tag = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &tag);

			if (tag == -1)
				tag = get_free_tag();

			/*
			for (DWORD s = 0; s < map.n_sectors; s++)
			{
				if (SELECTED(sectors, s))
					map.sectors[s]->tag = tag;

				if (!select && HILIGHTED(sectors, s))
					map.sectors[s]->tag = tag;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.sectors[selected_items.numbers[a]]->tag = tag;
			}
			else if (hilight_item != -1)
				map.sectors[hilight_item]->tag = tag;
		}
	}

	if (open_menu == MENU_LINETYPEGROUPS)
	{
		if (sel_item == line_groups.n_strings - 1)
		{
			bool select = selection();
			int type = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &type);

			if (type < 0)
				type = 0;

			/*
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				if (SELECTED(lines, l))
					map.lines[l]->type = type;

				if (!select && HILIGHTED(lines, l))
					map.lines[l]->type = type;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.lines[selected_items.numbers[a]]->type = type;
			}
			else if (hilight_item != -1)
				map.lines[hilight_item]->type = type;
		}
	}

	if (open_menu == MENU_THINGTYPEGROUPS)
	{
		if (sel_item == thing_groups.n_strings - 1)
		{
			bool select = selection();
			int type = atoi(menu_input.c_str());
			//sscanf(menu_input.c_str(), "%d", &type);

			if (type < 0)
				type = 0;

			/*
			for (DWORD t = 0; t < map.n_things; t++)
			{
				if (SELECTED(things, t))
					map.things[t]->type = type;

				if (!select && HILIGHTED(things, t))
					map.things[t]->type = type;
			}
			*/
			if (selection())
			{
				for (DWORD a = 0; a < selected_items.n_numbers; a++)
					map.things[selected_items.numbers[a]]->type = type;
			}
			else if (hilight_item != -1)
				map.things[hilight_item]->type = type;
		}
	}

	if (open_menu == MENU_THINGARGS)
	{
		bool select = selection();
		int arg = atoi(menu_input.c_str());

		/*
		for (DWORD t = 0; t < map.n_things; t++)
		{
			if (SELECTED(things, t))
				map.things[t]->args[sel_item] = arg;

			if (!select && HILIGHTED(things, t))
				map.things[t]->args[sel_item] = arg;
		}
		*/
		if (selection())
		{
			for (DWORD a = 0; a < selected_items.n_numbers; a++)
				map.things[selected_items.numbers[a]]->args[sel_item] = arg;
		}
		else if (hilight_item != -1)
			map.things[hilight_item]->args[sel_item] = arg;
	}

	if (open_menu == MENU_FILEMENU)
	{
		if (sel_item == 1)
		{
			//map.save_as(menu_input + ".wad", map.name);
			Wad *newwad = new Wad();
			newwad->path = menu_input + ".wad";
			map.add_to_wad(newwad);
			newwad->save(true);

			//delete(newwad);
		}

		menu = false;
	}

	if (open_menu == MENU_EDITMENU)
	{
		if (sel_item == 1)
		{
			if (edit_mode == 1)
			{
				int amount = atoi(menu_input.c_str());
				line_extrude(amount);
			}

			if (edit_mode == 2)
			{
				int height = atoi(menu_input.c_str());
				sector_create_stairs(height, false);
			}
		}

		if (sel_item == 2)
		{
			if (edit_mode == 2)
			{
				int height = atoi(menu_input.c_str());
				sector_create_stairs(height, true);
			}
		}
	}

	if (open_menu == MENU_DISPLAY)
	{
		if (sel_item == 3)
		{
			line_size = (float)atof(menu_input.c_str());
			update_map();
		}
	}

		// Browser menu
	if (open_menu == MENU_BROWSER)
	{
		if (sel_item == 0)
			browser_columns = atoi(menu_input.c_str());

		if (browser_columns < 4)
			browser_columns = 4;
	}
}

// menubar_click: When the user clicks on the menu bar
// ------------------------------------------------ >>
void menubar_click(int xpos)
{
	// File menu
	if (xpos > 8 && xpos < 40)
	{
		menu = true;
		sel_item = 0;
		max_item = file_menu.n_strings - 1;
		open_menu = MENU_FILEMENU;
	}

	// Edit menu
	if (xpos > 48 && xpos < 80)
	{
		menu = true;
		sel_item = 0;
		setup_edit_menu();
		max_item = edit_menu.n_strings - 1;
		open_menu = MENU_EDITMENU;
	}

	// Options menu
	if (xpos > 88 && xpos < 144)
	{
		menu = true;
		sel_item = 0;
		max_item = options_menu.n_strings - 1;
		open_menu = MENU_OPTIONSMENU;
	}

	// Mode menu
	if (xpos > 152 && xpos < 184)
	{
		menu = true;
		sel_item = 0;
		max_item = mode_menu.n_strings - 1;
		open_menu = MENU_MODEMENU;
	}

	// Check menu
	if (xpos > 192 && xpos < 232)
	{
		menu = true;
		sel_item = 0;
		max_item = check_menu.n_strings - 1;
		open_menu = MENU_CHECKMENU;
	}
}

bool check_mouse_menu()
{
	if (mouse.x < menu_area.x2() && mouse.x > menu_area.x1())
	{
		if (mouse.y < menu_area.y2() && mouse.y > menu_area.y1())
		{
			int y_off = mouse.y - menu_area.y1();
			sel_item = (int)(y_off / 10);
			return true;
		}
	}

	return false;
}

void draw_message()
{
	if (message)
		gui_popupmessage(popup_message, message_title.c_str());
}
