// <<--------------------------------------------->>
// << SLADE (SlayeR's 'LeetAss Doom Editor)       >>
// << By Simon Judd, 2004-05                      >>
// << ------------------------------------------- >>
// << <text_ui.cpp>                               >>
// << Functions dealing with the text-based ui    >>
// <<--------------------------------------------->>

// Includes ------------------------------------- >>
#include "main.h"
#include "map.h"

// External Variables --------------------------- >>
extern WadList wads;
extern stringlist_t search_paths;
extern Map map;
extern string game;
extern bool running;

EXTERN_CVAR(Int, vid_width)
EXTERN_CVAR(Int, vid_height)

// text_input: Gets user input and deals with it accordingly
// ------------------------------------------------------ >>
bool text_input()
{
	char in[256] = "";
	string input;

	printf("> ");
	fgets(in, 256, stdin);
	input = in;

	if (input.substr(0, 4) == "quit" || input.substr(0, 2) == "q\n")
	{
		running = false;
		return true;
	}

	if (input.substr(0, 8) == "listwads" || input.substr(0, 2) == "l\n")
	{
		wads.list_wads();
		return false;
	}

	if (input.substr(0, 4) == "vid " || input.substr(0, 2) == "v ")
	{
		int w = 0;
		int h = 0;

		sscanf(input.c_str(), "%s%d%d", &in, &w, &h);

		if (w != 0 && h != 0)
		{
			vid_width = w;
			vid_height = h;
			printf("Resolution set to %dx%d\n", w, h);
		}

		return false;
	}

	if (input.substr(0, 2) == "?\n")
	{
		printf("\nAvailable commands:\n");
		printf("-------------------\n");
		printf("(o)pen <path>          -- Opens a PWAD file\n");
		printf("(c)lose <path>         -- Closes a PWAD file (if open)\n");
		printf("close(a)ll             -- Closes all open PWADs\n");
		printf("(l)istwads             -- List open PWADs\n");
		printf("list(m)aps             -- List available maps for editing\n");
		printf("(e)dit <map>           -- Opens map for editing (in last wad opened)\n");
		printf("(b)uild <map>          -- Opens an empty map\n");
		printf("(v)id <width> <height> -- Set video mode\n");
		printf("(q)uit                 -- Quit SLADE\n");
		return false;
	}

	if (input.substr(0, 5) == "open " || input.substr(0, 2) == "o ")
	{
		string wadpath;
		sscanf(input.c_str(), "%s%s", &in, &in);
		wadpath = in;

		if (wadpath.find(".wad") == string::npos)
			wadpath += ".wad";

		if (!wads.open_wad(wadpath))
		{
			for (DWORD s = 0; s < search_paths.n_strings; s++)
			{
				string path = *search_paths.strings[s];

				if (path[path.length()] != '\\')
					path += '\\';

				path += wadpath;
				//printf("%s\n", path.c_str());

				if (wads.open_wad(path))
					return false;
			}

			printf("Wad %s cannot be found.\n", wadpath.c_str());
		}

		return false;
	}

	if (input.substr(0, 6) == "close " || input.substr(0, 2) == "c ")
	{
		string wadpath;
		sscanf(input.c_str(), "%s%s", &in, &in);
		wadpath = in;

		if (wadpath.find(".wad") == string::npos)
			wadpath += ".wad";

		if (!wads.close_wad(wadpath))
		{
			for (DWORD s = 0; s < search_paths.n_strings; s++)
			{
				string path = *search_paths.strings[s];

				if (path[path.length()] != '\\')
					path += '\\';

				path += wadpath;

				if (wads.close_wad(path))
					return false;
			}

			printf("Wad %s is not open.\n", wadpath.c_str());
		}

		return false;
	}

	if (input.substr(0, 5) == "edit " || input.substr(0, 2) == "e ")
	{
		string mapname;
		sscanf(input.c_str(), "%s%s", &in, &in);
		mapname = strupr(in);

		if (wads.get_last())
		{
			if (map.open(wads.get_last_withmaps(), mapname.c_str()))
				return true;
		}
		else
		{
			if (map.open(wads.get_iwad(), mapname.c_str()))
				return true;
		}

		return false;
	}

	if (input.substr(0, 6) == "build " || input.substr(0, 2) == "b ")
	{
		string mapname;
		sscanf(input.c_str(), "%s%s", &in, &in);
		mapname = strupr(in);

		map.create(mapname);
		return true;
	}

	if (input.substr(0, 8) == "listmaps" || input.substr(0, 2) == "m\n")
	{
		Wad* wadfile = wads.get_last_withmaps();

		if (!wadfile)
			wadfile = wads.get_iwad();

		if (wadfile)
		{
			if (wadfile->available_maps.n_strings == 0)
				printf("No available maps.\n");
			else
			{
				for (int s = 0; s < wadfile->available_maps.n_strings; s++)
				{
					if (s < wadfile->available_maps.n_strings - 1) 
						printf("%s, ", wadfile->available_maps.strings[s]->c_str());
					else
						printf("%s\n", wadfile->available_maps.strings[s]->c_str());
				}
			}
		}
		else
			printf("No wads open.\n");
	}

	if (input.substr(0, 8) == "closeall" || input.substr(0, 2) == "a\n")
	{
		wads.close_all();
		return false;
	}

	return false;
}

// text_welcome: Displays the welcome message
// --------------------------------------- >>
void text_welcome()
{
	FILE *fp = fopen("slade.log", "wt");
	fprintf(fp, "<< ----- SLADE LOGFILE ----- >>\n\n");
	fclose(fp);

	printf("<< ------------------------------------------------------------------ >>\n");
	printf("<< S.L.A.D.E. -- \"SlayeR's LeetAss Doom Editor\" (v0.5)                >>\n");
	printf("<< ------------------------------------------------------------------ >>\n");
	printf("<< By Simon \"SlayeR\" Judd, 2004                                       >>\n");
	printf("<< web: \"http://slade.mancubus.net\" email: \"slayer@mancubus.net\"      >>\n");
	printf("<< ------------------------------------------------------------------ >>\n");
	printf("<< Type '?' for a list of valid commands.                             >>\n");
	printf("<< ------------------------------------------------------------------ >>\n\n");
	print(false, "Using Game Configuration: \"%s\"\n", game.c_str());
	print(false, "Game IWAD: \"%s\"\n\n", wads.get_iwad()->path.c_str());
	//printf("%d", VK_RETURN);
}

// text_ui: The text-based ui main loop
// --------------------------------- >>
void text_ui()
{
	while (1)
	{
		if (text_input())
			break;
	}
}

void log_message(char *message)
{
	FILE *fp = fopen("slade.log", "a");
	fprintf(fp, message);
	fclose(fp);
}

void print(bool debug, char *message, ...)
{
	char string[512];
	va_list ap;

	va_start(ap, message);
	vsprintf(string, message, ap);
	va_end(ap);

	if (!debug)
		printf(string);

	log_message(string);
}
