
struct texture_t
{
	// Texture name
	string	name;

	// OpenGL Texture ID
	GLuint	tex_id;

	// Image dimensions
	WORD	width;
	WORD	height;

	// For non-power-of-two textures
	float	tex_x;	// Right edge GL tex-coord
	float	tex_y;	// Bottom edge GL tex-coord
	WORD	rwidth;
	WORD	rheight;

	// Temporary image data
	BYTE	*data;

	// Editing stuff
	bool	available;	// Texture is available for use in a map
	bool	flat;		// Texture is a flat
	//bool	loaded;		// Texture is loaded into memory

	// ZDoomGL Hires textures
	float	x_scale;
	float	y_scale;

	texture_t()
	{
		tex_id = width = height = 0;
		tex_x = tex_y = 1.0f;
		data = (BYTE *)NULL;
	}

	void set_data_size(WORD w, WORD h)
	{
		data = (BYTE *)malloc((w * h) * 4);
	}

	void clear_data()
	{
		for (int p = 0; p < rwidth * rheight * 4; p++)
			data[p] = 0;
	}

	void add_pixel(int x, int y, rgba_t colour)
	{
		if (x < 0 || x >= rwidth)
			return;

		if (y < 0 || y >= rheight)
			return;

		DWORD p = ((y * (rwidth)) + x) * 4;

		data[p] = colour.r;
		data[p + 1] = colour.g;
		data[p + 2] = colour.b;
		data[p + 3] = colour.a;
	}
};

class TextureList
{
private:
	texture_t	**textures;
	DWORD		n_textures;

public:
	TextureList() { textures = (texture_t **)NULL; n_textures = 0; }
	~TextureList() { free(textures); }

	DWORD		length() { return n_textures; }
	texture_t*	texture(DWORD index) { return textures[index]; }

	bool	add_file(string name, string filename, int filter, BYTE flags, bool available, bool flat);
	void	add_tex(string name, WORD width, WORD height, bool available, bool flat);
	void	init_tex_data();
	void	create_tex(int filter, BYTE flags);

	void	delete_tex(string name, bool available, bool flat);
	
	DWORD		get_tex_index(string name);
	DWORD		get_tex_index(string name, bool available, bool flat);
	texture_t*	get_tex(string name);
	texture_t*	get_tex(string name, bool available, bool flat);
	texture_t*	get_last();
	GLuint		gl_tex(string name);
	bool		tex_exists(string name, bool available, bool flat);
	string		get_name(GLuint id);
};

#define TEX_CLAPMX	0x01
#define TEX_CLAMPY	0x02