#include "main.h"
#include "map.h"

thinggroup_t	tgroups[32];
BYTE			n_tgroups = 0;
int				c_tgroup = 0;

// Defaults
rgba_t	def_col(255, 255, 255, 255);
WORD	def_rad = 20;
bool	angle = true;
bool	d_hanging = false;

extern Map map;

extern stringlist_t thing_groups;
extern stringlist_t thing_types[32];
extern stringlist_t sprites;
extern TextureList tex_list;

thingdef_t thing_unknown;

thing_t::thing_t()
{
	x = 0;
	y = 0;
	z = 0;
	type = 1;
	angle = 0;
	flags = THING_EASY|THING_MEDIUM|THING_HARD|THING_DEAF|THING_SINGLE|THING_COOPERATIVE|THING_FIGHTER|THING_CLERIC|THING_MAGE;
	tid = 0;
	special = 0;
	draw_angle = false;

	for (int a = 0; a < 5; a++)
		args[a] = 0;

	colour.set(255, 255, 255, 255);
	radius = 20;

	//hilighted = selected = moving = false;
}

thingdef_t::thingdef_t()
{
	type = 0;
	name = "Unknown";
	colour.set(def_col);
	size = def_rad;
	args[0] = "Arg1";
	args[1] = "Arg2";
	args[2] = "Arg3";
	args[3] = "Arg4";
	args[4] = "Arg5";
	show_angle = angle;
	spritename = "_unknownsprite";
	hanging = d_hanging;

	for (int a = 0; a < 5; a++)
		arg_types[a] = NULL;
}

void parse_rgb(MemReader *mr, rgba_t* col)
{
	// Read opening brace
	mr->check_token("{");

	// Read rgb
	col->r = mr->get_integer();
	col->g = mr->get_integer();
	col->b = mr->get_integer();
	col->a = 255;

	// Read closing brace
	mr->check_token("}");
}

void parse_thing(MemReader *mr)
{
	thingdef_t *tdef = new thingdef_t;

	// Read thing type
	tdef->type = mr->get_integer();

	// Read thing name
	tdef->name = mr->get_token();
	//printf("Added thing \"%s\"\n", tdef->name.c_str());

	// Check for extra info
	if (mr->peek_token() == "{")
	{
		string token = mr->get_token();

		while (token != "}")
		{
			if (token == "colour")
				parse_rgb(mr, &tdef->colour);

			if (token == "size")
				tdef->size = mr->get_integer();

			if (token == "arg1")
				tdef->args[0] = mr->get_token();

			if (token == "arg2")
				tdef->args[1] = mr->get_token();

			if (token == "arg3")
				tdef->args[2] = mr->get_token();

			if (token == "arg4")
				tdef->args[3] = mr->get_token();

			if (token == "arg5")
				tdef->args[4] = mr->get_token();

			if (token == "sprite")
			{
				tdef->spritename = mr->get_token();

				if (tdef->spritename[0] != '_')
					sprites.add(tdef->spritename);
					//sprites.add(tdef->sprite);
				//printf("Added sprite \"%s\"\n", tdef->sprite.c_str());
			}

			if (token == "hanging")
				tdef->hanging = mr->get_bool();

			//tdef->dump();

			token = mr->get_token();
		}
	}

	//tdef->dump();
	tgroups[c_tgroup].add(tdef);
	thing_types[c_tgroup].add(tdef->name);
}

int get_tgroup(string name)
{
	int group = -1;

	for (BYTE g = 0; g < n_tgroups; g++)
	{
		if (tgroups[g].name == name)
			group = g;
	}

	return group;
}

void parse_thing_group(MemReader *mr)
{
	string	token;
	d_hanging = false;

	// Read group name
	token = mr->get_token();
	c_tgroup = get_tgroup(token);

	if (c_tgroup < 0)
	{
		tgroups[n_tgroups].name = token;
		c_tgroup = n_tgroups;
		n_tgroups++;
		thing_groups.add(token);
		//printf("Added group \"%s\"\n", tgroups[c_tgroup].name.c_str());
	}
	//else
		//printf("Add to group \"%s\"\n", tgroups[c_tgroup].name.c_str());

	// Read opening brace
	if (mr->get_token() != "{")
	{
		message_box("Error reading thing group \"" + tgroups[c_tgroup].name + "\": { expected.");
		return;
	}

	while (token != "}")
	{
		if (token == "default_colour")
			parse_rgb(mr, &def_col);

		if (token == "default_size")
			def_rad = mr->get_integer();

		if (token == "angle")
			angle = mr->get_bool();

		if (token == "hanging")
			d_hanging = mr->get_bool();

		if (token == "thing")
			parse_thing(mr);

		token = mr->get_token();
	}
}

thingdef_t* get_tdef(short type)
{
	for (DWORD g = 0; g < n_tgroups; g++)
	{
		for (DWORD t = 0; t < tgroups[g].n_defs; t++)
		{
			if (tgroups[g].defs[t]->type == type)
				return tgroups[g].defs[t];
		}
	}

	return &thing_unknown;
}

thingdef_t* get_tdef(string name)
{
	for (DWORD g = 0; g < n_tgroups; g++)
	{
		for (DWORD t = 0; t < tgroups[g].n_defs; t++)
		{
			if (tgroups[g].defs[t]->name == name)
				return tgroups[g].defs[t];
		}
	}

	return &thing_unknown;
}

void update_tdef_textures()
{
	for (DWORD g = 0; g < n_tgroups; g++)
	{
		for (DWORD t = 0; t < tgroups[g].n_defs; t++)
			tgroups[g].defs[t]->sprite = tex_list.get_tex(tgroups[g].defs[t]->spritename, false, false);
	}

	thing_unknown.sprite = tex_list.get_tex(thing_unknown.spritename, false, false);
}

void thing_settype(BYTE group, WORD item, DWORD thing)
{
	map.things[thing]->type = tgroups[group].defs[item]->type;
}

void update_map_things()
{
	thing_unknown.size = -1;

	thingdef_t *tdef = NULL;

	for (DWORD t = 0; t < map.n_things; t++)
	{
		tdef = get_tdef(map.things[t]->type);

		map.things[t]->colour.set(tdef->colour);
		map.things[t]->radius = tdef->size;
		map.things[t]->draw_angle = tdef->show_angle;
		map.things[t]->tdef = tdef;
	}
}

void dump_thing_names(stringlist_t *list)
{
	for (DWORD g = 0; g < n_tgroups; g++)
	{
		for (DWORD t = 0; t < tgroups[g].n_defs; t++)
			list->add(tgroups[g].defs[t]->name);
	}
}